"""
PKCS#11 DSA Tests
"""

import base64

import pkcs11
from pkcs11 import Attribute, KeyType, Mechanism
from pkcs11.util.dsa import (
    decode_dsa_domain_parameters,
    encode_dsa_domain_parameters,
)

from . import FIXME, TestCase, requires

DHPARAMS = base64.b64decode("""
MIIBHwKBgQD8jXSat2sk+j0plaMn51AVYBWEyWee3ui3llRUckVceDILsjVdBs1tXCDhU7WC+VZZ
u6ujBHZONiXcQTZ6P/jhnYlSyjEoBTf7GntlbjeASm63XYzTt4E5i7u1RI6TmEIRj6VTrM5m5DFP
fDQ+fflAJzm0phT38gYE5xfe3mmCDQIVAMIMNr/4lufeH46EGKQXVnvtJBAZAoGBANxCIKAfh1/v
MvI/2s7S1ESGuwvmvbFWpxW3gNXvyO2mWjfHC3sQrwm3qED0R71n9bIL6VqRK+tBEy6VkR+lKifA
8rPnZvADPNBhRLhgDc4JuwYinRJSUPd1iZxJCbumfscr3Fp1XuUnCcMRkWqWr7rGEUP+ht+AeXpo
ouQbj2Vq
""")


class DSATests(TestCase):
    @requires(Mechanism.DSA_PARAMETER_GEN)
    @FIXME.nfast  # returns Function Failed
    def test_generate_params(self):
        parameters = self.session.generate_domain_parameters(KeyType.DSA, 1024)
        self.assertIsInstance(parameters, pkcs11.DomainParameters)
        self.assertEqual(parameters[Attribute.PRIME_BITS], 1024)

        encode_dsa_domain_parameters(parameters)

    @requires(Mechanism.DSA_KEY_PAIR_GEN, Mechanism.DSA_SHA1)
    def test_generate_keypair_and_sign(self):
        dhparams = self.session.create_domain_parameters(
            KeyType.DSA, decode_dsa_domain_parameters(DHPARAMS), local=True
        )

        public, private = dhparams.generate_keypair()
        self.assertIsInstance(public, pkcs11.PublicKey)
        self.assertIsInstance(private, pkcs11.PrivateKey)
        # We expect a length of 128 (1024/8) in the vast majority of cases,
        # but since the length of an integer value in DER is not fixed, there's
        # a chance that we end up with a slightly shorter key length.
        # The probability that the length falls short of 120 is vanishingly low, though.
        self.assertGreater(len(public[Attribute.VALUE]), 120)

        data = "Message to sign"
        signature = private.sign(data, mechanism=Mechanism.DSA_SHA1)
        self.assertTrue(public.verify(data, signature, mechanism=Mechanism.DSA_SHA1))

    @requires(Mechanism.DSA_PARAMETER_GEN, Mechanism.DSA_KEY_PAIR_GEN)
    @FIXME.nfast  # returns Function Failed
    def test_generate_keypair_directly(self):
        public, private = self.session.generate_keypair(KeyType.DSA, 1024)
        # See above.
        self.assertGreater(len(public[Attribute.VALUE]), 120)



"""
PKCS#11 Tests

The following environment variables will influence the behaviour of test cases:
 - PKCS11_MODULE, mandatory, points to the library/DLL to use for testing
 - PKCS11_TOKEN_LABEL, mandatory, contains the token label
 - PKCS11_TOKEN_PIN, optional (default is None), contains the PIN/passphrase of the token
 - PKCS11_TOKEN_SO_PIN, optional (default is same as PKCS11_TOKEN_PIN), security officer PIN
 - OPENSSL_PATH, optional, path to openssl executable (i.e. the folder that contains it)

"""

import os
import shutil
import unittest
from functools import wraps
from warnings import warn

import pkcs11

try:
    LIB = os.environ["PKCS11_MODULE"]
except KeyError as ex:
    raise RuntimeError("Must define `PKCS11_MODULE' to run tests.") from ex


try:
    TOKEN = os.environ["PKCS11_TOKEN_LABEL"]
except KeyError as ex:
    raise RuntimeError("Must define `PKCS11_TOKEN_LABEL' to run tests.") from ex

TOKEN_PIN = os.environ.get("PKCS11_TOKEN_PIN")  # Can be None
if TOKEN_PIN is None:
    warn("`PKCS11_TOKEN_PIN' env variable is unset.", stacklevel=2)

TOKEN_SO_PIN = os.environ.get("PKCS11_TOKEN_SO_PIN")
if TOKEN_SO_PIN is None:
    TOKEN_SO_PIN = TOKEN_PIN
    warn(
        "`PKCS11_TOKEN_SO_PIN' env variable is unset. Using value from `PKCS11_TOKEN_PIN'",
        stacklevel=2,
    )

OPENSSL = shutil.which("openssl", path=os.environ.get("OPENSSL_PATH"))
if OPENSSL is None:
    warn("Path to OpenSSL not found. Please adjust `PATH' or define `OPENSSL_PATH'", stacklevel=2)


class TestCase(unittest.TestCase):
    """Base test case, optionally creates a token and a session."""

    with_token = True
    """Creates a token for this test case."""
    with_session = True
    """Creates a session for this test case."""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.lib = lib = pkcs11.lib(LIB)

        if cls.with_token or cls.with_session:
            cls.token = lib.get_token(token_label=TOKEN)

    def setUp(self):
        super().setUp()

        if self.with_session:
            self.session = self.token.open(user_pin=TOKEN_PIN)

    def tearDown(self):
        if self.with_session:
            self.session.close()

        super().tearDown()


def requires(*mechanisms):
    """
    Decorates a function or class as requiring mechanisms, else they are
    skipped.
    """

    def check_requirements(self):
        """Determine what, if any, required mechanisms are unavailable."""
        unavailable = set(mechanisms) - self.token.slot.get_mechanisms()

        if unavailable:
            raise unittest.SkipTest("Requires %s" % ", ".join(map(str, unavailable)))

    def inner(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            check_requirements(self)

            return func(self, *args, **kwargs)

        return wrapper

    return inner


def xfail(condition):
    """Mark a test that's expected to fail for a given condition."""

    def inner(func):
        if condition:
            return unittest.expectedFailure(func)

        else:
            return func

    return inner


class Is:
    """
    Test what device we're using.
    """

    # trick: str.endswith() can accept tuples,
    # see https://stackoverflow.com/questions/18351951/check-if-string-ends-with-one-of-the-strings-from-a-list
    softhsm2 = LIB.lower().endswith(
        ("libsofthsm2.so", "libsofthsm2.dylib", "softhsm2.dll", "softhsm2-x64.dll")
    )
    nfast = LIB.lower().endswith(("libcknfast.so", "cknfast.dll"))
    opencryptoki = LIB.endswith("libopencryptoki.so")
    travis = os.environ.get("TRAVIS") == "true"


class Avail:
    """
    Test if a resource is available
    """

    # openssl is searched across the exec path. Optionally, OPENSSL_PATH env variable can be defined
    # in case there is no direct path to it (i.e. PATH does not point to it)
    openssl = OPENSSL is not None


class Only:
    """
    Limit tests to given conditions
    """

    softhsm2 = unittest.skipUnless(Is.softhsm2, "SoftHSMv2 only")
    openssl = unittest.skipUnless(Avail.openssl, "openssl not found in the path")


class Not:
    """
    Ignore tests for given devices
    """

    softhsm2 = unittest.skipIf(Is.softhsm2, "Not supported by SoftHSMv2")
    nfast = unittest.skipIf(Is.nfast, "Not supported by nFast")
    opencryptoki = unittest.skipIf(Is.opencryptoki, "Not supported by OpenCryptoki")


class FIXME:
    """
    Tests is broken on this platform.
    """

    softhsm2 = xfail(Is.softhsm2)
    nfast = xfail(Is.nfast)
    opencryptoki = xfail(Is.opencryptoki)
    travis = xfail(Is.travis)



"""
PKCS#11 Slots and Tokens
"""

import os
import unittest

import pkcs11

from . import LIB


@unittest.skipUnless("PKCS11_MODULE2" in os.environ, "Requires an additional PKCS#11 module")
class MultilibTests(unittest.TestCase):
    def test_double_initialise_different_libs(self):
        lib1 = pkcs11.lib(LIB)
        lib2 = pkcs11.lib(os.environ["PKCS11_MODULE2"])
        self.assertIsNotNone(lib1)
        self.assertIsNotNone(lib2)
        self.assertIsNot(lib1, lib2)

        slots1 = lib1.get_slots()
        slots2 = lib2.get_slots()

        self.assertGreaterEqual(len(slots1), 1)
        self.assertGreaterEqual(len(slots2), 1)



"""
PKCS#11 Slots and Tokens
"""

import unittest

import pkcs11
from pkcs11 import PKCS11Error

from . import LIB, TOKEN, Not, Only


class SlotsAndTokensTests(unittest.TestCase):
    def test_double_initialise(self):
        attempt1 = pkcs11.lib(LIB)
        attempt2 = pkcs11.lib(LIB)
        self.assertIsNotNone(attempt1)
        self.assertIsNotNone(attempt2)
        self.assertIs(attempt1, attempt2)

    def test_nonexistent_lib(self):
        with self.assertRaises(RuntimeError):
            pkcs11.lib("thislibdoesntexist.so")

    def test_double_initialise_nonexistent_lib(self):
        self.assertIsNotNone(pkcs11.lib(LIB))
        with self.assertRaises(RuntimeError):
            pkcs11.lib("somethingelse.so")

    @Only.softhsm2
    def test_get_slots(self):
        lib = pkcs11.lib(LIB)
        slots = lib.get_slots()

        self.assertEqual(len(slots), 2)
        slot1, slot2 = slots

        self.assertIsInstance(slot1, pkcs11.Slot)
        self.assertEqual(slot1.flags, pkcs11.SlotFlag.TOKEN_PRESENT)

    def test_get_mechanisms(self):
        lib = pkcs11.lib(LIB)
        slot, *_ = lib.get_slots()
        mechanisms = slot.get_mechanisms()
        self.assertIn(pkcs11.Mechanism.RSA_PKCS, mechanisms)

    def test_reinitialize(self):
        lib = pkcs11.lib(LIB)
        slots = lib.get_slots()
        self.assertGreaterEqual(len(slots), 1)

        lib.reinitialize()

        self.assertTrue(lib.initialized)
        lib = pkcs11.lib(LIB)
        slots = lib.get_slots()
        self.assertGreaterEqual(len(slots), 1)

    def test_finalize(self):
        lib = pkcs11.lib(LIB)
        slots = lib.get_slots()
        self.assertGreaterEqual(len(slots), 1)

        lib.finalize()
        self.assertFalse(lib.initialized)
        self.assertRaises(PKCS11Error, lib.get_slots)

    def test_auto_reinitialise(self):
        lib = pkcs11.lib(LIB)
        lib.finalize()
        self.assertFalse(lib.initialized)
        lib = pkcs11.lib(LIB)
        slots = lib.get_slots()
        self.assertGreaterEqual(len(slots), 1)

    def test_unload_reload(self):
        pkcs11.lib(LIB)
        pkcs11.unload(LIB)

        lib = pkcs11.lib(LIB)
        slots = lib.get_slots()
        self.assertGreaterEqual(len(slots), 1)

    def test_get_mechanism_info(self):
        lib = pkcs11.lib(LIB)
        slot, *_ = lib.get_slots()
        info = slot.get_mechanism_info(pkcs11.Mechanism.RSA_PKCS_OAEP)
        self.assertIsInstance(info, pkcs11.MechanismInfo)

    @Not.nfast  # EC not supported
    @Not.opencryptoki
    def test_get_mechanism_info_ec(self):
        lib = pkcs11.lib(LIB)
        slot, *_ = lib.get_slots()
        info = slot.get_mechanism_info(pkcs11.Mechanism.EC_KEY_PAIR_GEN)
        self.assertIsInstance(info, pkcs11.MechanismInfo)
        self.assertIn(pkcs11.MechanismFlag.EC_NAMEDCURVE, info.flags)

    @Only.softhsm2
    def test_get_tokens(self):
        lib = pkcs11.lib(LIB)

        tokens = lib.get_tokens(token_flags=pkcs11.TokenFlag.RNG)
        self.assertEqual(len(list(tokens)), 2)

        tokens = lib.get_tokens(token_label=TOKEN)
        self.assertEqual(len(list(tokens)), 1)

    @Only.softhsm2
    def test_get_token(self):
        lib = pkcs11.lib(LIB)
        slot, *_ = lib.get_slots()
        token = slot.get_token()

        self.assertIsInstance(token, pkcs11.Token)
        self.assertEqual(token.label, TOKEN)
        self.assertIn(pkcs11.TokenFlag.TOKEN_INITIALIZED, token.flags)
        self.assertIn(pkcs11.TokenFlag.LOGIN_REQUIRED, token.flags)



"""
PKCS#11 DES Secret Keys
"""

import pkcs11
from pkcs11 import KeyType, Mechanism

from . import TestCase, requires


class DESTests(TestCase):
    @requires(Mechanism.DES2_KEY_GEN)
    def test_generate_des2_key(self):
        key = self.session.generate_key(KeyType.DES2)
        self.assertIsInstance(key, pkcs11.SecretKey)

    @requires(Mechanism.DES3_KEY_GEN)
    def test_generate_des3_key(self):
        key = self.session.generate_key(KeyType.DES3)
        self.assertIsInstance(key, pkcs11.SecretKey)

    @requires(Mechanism.DES2_KEY_GEN, Mechanism.DES3_CBC_PAD)
    def test_encrypt_des2(self):
        key = self.session.generate_key(KeyType.DES2)

        iv = self.session.generate_random(64)
        crypttext = key.encrypt("PLAIN TEXT_", mechanism_param=iv)
        plaintext = key.decrypt(crypttext, mechanism_param=iv)

        self.assertEqual(plaintext, b"PLAIN TEXT_")

    @requires(Mechanism.DES3_KEY_GEN, Mechanism.DES3_CBC_PAD)
    def test_encrypt_des3(self):
        key = self.session.generate_key(KeyType.DES3)

        iv = self.session.generate_random(64)
        crypttext = key.encrypt("PLAIN TEXT_", mechanism_param=iv)
        plaintext = key.decrypt(crypttext, mechanism_param=iv)

        self.assertEqual(plaintext, b"PLAIN TEXT_")



"""
PKCS#11 RSA Public Key Cryptography
"""

import pkcs11
from pkcs11 import MGF, Attribute, KeyType, Mechanism, ObjectClass

from . import FIXME, TOKEN_PIN, TestCase, requires


class RSATests(TestCase):
    @requires(Mechanism.RSA_PKCS_KEY_PAIR_GEN)
    def setUp(self):
        super().setUp()

        self.public, self.private = self.session.generate_keypair(KeyType.RSA, 1024)

    def test_key_length(self):
        self.assertEqual(1024, self.private.key_length)
        self.assertEqual(1024, self.public.key_length)

    @requires(Mechanism.RSA_PKCS)
    def test_sign_pkcs_v15(self):
        data = b"00000000"

        signature = self.private.sign(data, mechanism=Mechanism.RSA_PKCS)
        self.assertIsNotNone(signature)
        self.assertIsInstance(signature, bytes)
        self.assertTrue(self.public.verify(data, signature, mechanism=Mechanism.RSA_PKCS))
        self.assertFalse(self.public.verify(data, b"1234", mechanism=Mechanism.RSA_PKCS))

    @requires(Mechanism.SHA512_RSA_PKCS)
    def test_sign_with_reauthentication(self):
        public, private = self.session.generate_keypair(
            KeyType.RSA, 1024, private_template={Attribute.ALWAYS_AUTHENTICATE: True}
        )
        data = "INPUT"

        signature = private.sign(data, pin=TOKEN_PIN)
        self.assertIsNotNone(signature)
        self.assertIsInstance(signature, bytes)
        self.assertTrue(public.verify(data, signature))

    @requires(Mechanism.SHA512_RSA_PKCS)
    def test_sign_default(self):
        data = b"HELLO WORLD" * 1024

        signature = self.private.sign(data)
        self.assertIsNotNone(signature)
        self.assertIsInstance(signature, bytes)
        self.assertTrue(self.public.verify(data, signature))
        self.assertFalse(self.public.verify(data, b"1234"))

    @requires(Mechanism.SHA512_RSA_PKCS)
    def test_sign_stream(self):
        data = (
            b"I" * 16,
            b"N" * 16,
            b"P" * 16,
            b"U" * 16,
            b"T" * 10,
        )

        signature = self.private.sign(data)
        self.assertIsNotNone(signature)
        self.assertIsInstance(signature, bytes)
        self.assertTrue(self.public.verify(data, signature))

    @requires(Mechanism.SHA512_RSA_PKCS)
    def test_sign_stream_with_reauthentication(self):
        public, private = self.session.generate_keypair(
            KeyType.RSA, 1024, private_template={Attribute.ALWAYS_AUTHENTICATE: True}
        )
        data = (
            b"I" * 16,
            b"N" * 16,
            b"P" * 16,
            b"U" * 16,
            b"T" * 10,
        )

        signature = private.sign(data, pin=TOKEN_PIN)
        self.assertIsNotNone(signature)
        self.assertIsInstance(signature, bytes)
        self.assertTrue(public.verify(data, signature))

    @requires(Mechanism.SHA512_RSA_PKCS)
    def test_sign_stream_with_empty_blocks(self):
        data = (
            b"I" * 16,
            b"N" * 16,
            b"",
            b"P" * 16,
            b"" * 10,
            b"U" * 16,
            b"T" * 10,
        )

        signature = self.private.sign(data)
        self.assertIsNotNone(signature)
        self.assertIsInstance(signature, bytes)
        self.assertTrue(self.public.verify(data, signature))

    @requires(Mechanism.SHA512_RSA_PKCS)
    def test_sign_stream_undersized_buffer(self):
        data = (
            b"I" * 16,
            b"N" * 16,
            b"P" * 16,
            b"U" * 16,
            b"T" * 10,
        )

        signature = self.private.sign(data, buffer_size=16)
        self.assertIsNotNone(signature)
        self.assertIsInstance(signature, bytes)
        self.assertTrue(self.public.verify(data, signature))

    @requires(Mechanism.SHA512_RSA_PKCS)
    def test_sign_stream_interrupt_releases_operation(self):
        data = (
            b"I" * 16,
            b"N" * 16,
            b"P" * 16,
            b"U" * 16,
            b"T" * 10,
        )

        def _data_with_error():
            yield data[0]
            yield data[1]
            yield data[2]
            raise ValueError

        def attempt_sign():
            self.private.sign(_data_with_error())

        self.assertRaises(ValueError, attempt_sign)
        # ...try again
        signature = self.private.sign(data)
        self.assertIsNotNone(signature)
        self.assertIsInstance(signature, bytes)
        self.assertTrue(self.public.verify(data, signature))

    @requires(Mechanism.SHA512_RSA_PKCS)
    def test_verify_stream_interrupt_releases_operation(self):
        data = (
            b"I" * 16,
            b"N" * 16,
            b"P" * 16,
            b"U" * 16,
            b"T" * 10,
        )

        def _data_with_error():
            yield data[0]
            yield data[1]
            yield data[2]
            raise ValueError

        signature = self.private.sign(data)

        def attempt_verify():
            self.public.verify(_data_with_error(), signature)

        self.assertRaises(ValueError, attempt_verify)
        # ...try again
        self.assertTrue(self.public.verify(data, signature))

    @requires(Mechanism.RSA_PKCS_OAEP)
    @FIXME.opencryptoki  # can't set key attributes
    def test_key_wrap(self):
        key = self.session.generate_key(
            KeyType.AES,
            128,
            template={
                Attribute.EXTRACTABLE: True,
                Attribute.SENSITIVE: False,
            },
        )

        data = self.public.wrap_key(key)
        self.assertNotEqual(data, key[Attribute.VALUE])

        key2 = self.private.unwrap_key(
            ObjectClass.SECRET_KEY,
            KeyType.AES,
            data,
            template={
                Attribute.EXTRACTABLE: True,
                Attribute.SENSITIVE: False,
            },
        )

        self.assertEqual(key[Attribute.VALUE], key2[Attribute.VALUE])

    @requires(Mechanism.RSA_PKCS_OAEP)
    def test_encrypt_oaep(self):
        data = b"SOME DATA"

        crypttext = self.public.encrypt(
            data,
            mechanism=Mechanism.RSA_PKCS_OAEP,
            mechanism_param=(Mechanism.SHA_1, MGF.SHA1, None),
        )

        self.assertNotEqual(data, crypttext)

        plaintext = self.private.decrypt(
            crypttext,
            mechanism=Mechanism.RSA_PKCS_OAEP,
            mechanism_param=(Mechanism.SHA_1, MGF.SHA1, None),
        )

        self.assertEqual(data, plaintext)

    @requires(Mechanism.SHA1_RSA_PKCS_PSS)
    def test_sign_pss(self):
        data = b"SOME DATA"

        # These are the default params
        signature = self.private.sign(
            data,
            mechanism=Mechanism.SHA1_RSA_PKCS_PSS,
            mechanism_param=(Mechanism.SHA_1, MGF.SHA1, 20),
        )

        self.assertTrue(self.public.verify(data, signature, mechanism=Mechanism.SHA1_RSA_PKCS_PSS))

    @requires(Mechanism.SHA1_RSA_PKCS_PSS)
    def test_sign_pss_undersized_buffer(self):
        data = b"SOME DATA"

        signature = self.private.sign(
            data,
            mechanism=Mechanism.SHA1_RSA_PKCS_PSS,
            mechanism_param=(Mechanism.SHA_1, MGF.SHA1, 20),
            buffer_size=16,
        )

        self.assertTrue(self.public.verify(data, signature, mechanism=Mechanism.SHA1_RSA_PKCS_PSS))

    @requires(Mechanism.RSA_PKCS_OAEP)
    def test_encrypt_too_much_data(self):
        data = b"1234" * 128

        # You can't encrypt lots of data with RSA
        # This should ideally throw DataLen but you can't trust it
        with self.assertRaises(pkcs11.PKCS11Error):
            self.public.encrypt(data)



"""
PKCS#11 Diffie-Hellman tests
"""

import base64

from pkcs11 import Attribute, DomainParameters, KeyType, Mechanism
from pkcs11.util.dh import (
    decode_dh_domain_parameters,
    encode_dh_domain_parameters,
    encode_dh_public_key,
)

from . import FIXME, TestCase, requires


class DHTests(TestCase):
    @requires(Mechanism.DH_PKCS_KEY_PAIR_GEN, Mechanism.DH_PKCS_DERIVE)
    @FIXME.opencryptoki  # AttributeValueInvalid when generating keypair
    def test_derive_key(self):
        # Alice and Bob each create a Diffie-Hellman keypair from the
        # publicly available DH parameters
        #
        # E.g. RFC 3526, RFC 5114 or openssl dhparam -C 2236
        prime = [
            0x0F,
            0x52,
            0xE5,
            0x24,
            0xF5,
            0xFA,
            0x9D,
            0xDC,
            0xC6,
            0xAB,
            0xE6,
            0x04,
            0xE4,
            0x20,
            0x89,
            0x8A,
            0xB4,
            0xBF,
            0x27,
            0xB5,
            0x4A,
            0x95,
            0x57,
            0xA1,
            0x06,
            0xE7,
            0x30,
            0x73,
            0x83,
            0x5E,
            0xC9,
            0x23,
            0x11,
            0xED,
            0x42,
            0x45,
            0xAC,
            0x49,
            0xD3,
            0xE3,
            0xF3,
            0x34,
            0x73,
            0xC5,
            0x7D,
            0x00,
            0x3C,
            0x86,
            0x63,
            0x74,
            0xE0,
            0x75,
            0x97,
            0x84,
            0x1D,
            0x0B,
            0x11,
            0xDA,
            0x04,
            0xD0,
            0xFE,
            0x4F,
            0xB0,
            0x37,
            0xDF,
            0x57,
            0x22,
            0x2E,
            0x96,
            0x42,
            0xE0,
            0x7C,
            0xD7,
            0x5E,
            0x46,
            0x29,
            0xAF,
            0xB1,
            0xF4,
            0x81,
            0xAF,
            0xFC,
            0x9A,
            0xEF,
            0xFA,
            0x89,
            0x9E,
            0x0A,
            0xFB,
            0x16,
            0xE3,
            0x8F,
            0x01,
            0xA2,
            0xC8,
            0xDD,
            0xB4,
            0x47,
            0x12,
            0xF8,
            0x29,
            0x09,
            0x13,
            0x6E,
            0x9D,
            0xA8,
            0xF9,
            0x5D,
            0x08,
            0x00,
            0x3A,
            0x8C,
            0xA7,
            0xFF,
            0x6C,
            0xCF,
            0xE3,
            0x7C,
            0x3B,
            0x6B,
            0xB4,
            0x26,
            0xCC,
            0xDA,
            0x89,
            0x93,
            0x01,
            0x73,
            0xA8,
            0x55,
            0x3E,
            0x5B,
            0x77,
            0x25,
            0x8F,
            0x27,
            0xA3,
            0xF1,
            0xBF,
            0x7A,
            0x73,
            0x1F,
            0x85,
            0x96,
            0x0C,
            0x45,
            0x14,
            0xC1,
            0x06,
            0xB7,
            0x1C,
            0x75,
            0xAA,
            0x10,
            0xBC,
            0x86,
            0x98,
            0x75,
            0x44,
            0x70,
            0xD1,
            0x0F,
            0x20,
            0xF4,
            0xAC,
            0x4C,
            0xB3,
            0x88,
            0x16,
            0x1C,
            0x7E,
            0xA3,
            0x27,
            0xE4,
            0xAD,
            0xE1,
            0xA1,
            0x85,
            0x4F,
            0x1A,
            0x22,
            0x0D,
            0x05,
            0x42,
            0x73,
            0x69,
            0x45,
            0xC9,
            0x2F,
            0xF7,
            0xC2,
            0x48,
            0xE3,
            0xCE,
            0x9D,
            0x74,
            0x58,
            0x53,
            0xE7,
            0xA7,
            0x82,
            0x18,
            0xD9,
            0x3D,
            0xAF,
            0xAB,
            0x40,
            0x9F,
            0xAA,
            0x4C,
            0x78,
            0x0A,
            0xC3,
            0x24,
            0x2D,
            0xDB,
            0x12,
            0xA9,
            0x54,
            0xE5,
            0x47,
            0x87,
            0xAC,
            0x52,
            0xFE,
            0xE8,
            0x3D,
            0x0B,
            0x56,
            0xED,
            0x9C,
            0x9F,
            0xFF,
            0x39,
            0xE5,
            0xE5,
            0xBF,
            0x62,
            0x32,
            0x42,
            0x08,
            0xAE,
            0x6A,
            0xED,
            0x88,
            0x0E,
            0xB3,
            0x1A,
            0x4C,
            0xD3,
            0x08,
            0xE4,
            0xC4,
            0xAA,
            0x2C,
            0xCC,
            0xB1,
            0x37,
            0xA5,
            0xC1,
            0xA9,
            0x64,
            0x7E,
            0xEB,
            0xF9,
            0xD3,
            0xF5,
            0x15,
            0x28,
            0xFE,
            0x2E,
            0xE2,
            0x7F,
            0xFE,
            0xD9,
            0xB9,
            0x38,
            0x42,
            0x57,
            0x03,
        ]
        parameters = self.session.create_domain_parameters(
            KeyType.DH,
            {
                Attribute.PRIME: prime,
                Attribute.BASE: [0x2],
            },
            local=True,
        )

        # Alice generate a keypair
        alice_public, alice_private = parameters.generate_keypair()
        alice_value = alice_public[Attribute.VALUE]

        # Bob generates a keypair
        bob_public, bob_private = parameters.generate_keypair()
        bob_value = bob_public[Attribute.VALUE]

        self.assertNotEqual(alice_value, bob_value)

        # Alice and Bob exchange values and an IV ...
        iv = self.session.generate_random(128)

        alice_session = alice_private.derive_key(
            KeyType.AES,
            128,
            mechanism_param=bob_value,
            template={
                Attribute.SENSITIVE: False,
                Attribute.EXTRACTABLE: True,
            },
        )
        bob_session = bob_private.derive_key(
            KeyType.AES,
            128,
            mechanism_param=alice_value,
            template={
                Attribute.SENSITIVE: False,
                Attribute.EXTRACTABLE: True,
            },
        )

        self.assertEqual(alice_session[Attribute.VALUE], bob_session[Attribute.VALUE])

        crypttext = alice_session.encrypt("HI BOB!", mechanism_param=iv)
        plaintext = bob_session.decrypt(crypttext, mechanism_param=iv)
        self.assertEqual(plaintext, b"HI BOB!")

    def test_load_params(self):
        # This is RFC5114 #2
        PARAMS = base64.b64decode("""
        MIICKQKCAQEArRB+HpEjqdDWYPqnlVnFH6INZOVoO5/RtUsVl7YdCnXm+hQd+VpW
        26+aPEB7od8V6z1oijCcGA4d5rhaEnSgpm0/gVKtasISkDfJ7e/aTfjZHo/vVbc5
        S3rVt9C2wSIHyfmNEe002/bGugssi7wnvmoA4KC5xJcIs7+KMXCRiDaBKGEwvImF
        2xYC5xRBXZMwJ4Jzx94x79xzEPcSH9WgdBWYfZrcCkhtzfk6zEQyg4cxXXXhmMZB
        pIDNhqG55YfovmDmnMkosrnFIXLkEwQumyPxCw4W55djybU9z0uoCinj+3PBa451
        uX7zY+L/ox9xz53lOE5xuBwKxN/+DBDmTwKCAQEArEAy708tmuOd8wtcj/2sUGze
        vnuJmYyvdIZqCM/k/+OmgkpOELmm8N2SHwGnDEr6q3OddwDCn1LFfbF8YgqGUr5e
        kAGo1mrXwXZpEBmZAkr00CcnWsE0i7inYtBSG8mK4kcVBCLqHtQJk51U2nRgzbX2
        xrJQcXy+8YDrNBGOmNEZUppF1vg0Vm4wJeMWozDvu3eobwwasVsFGuPUKMj4rLcK
        gTcVC47rEOGD7dGZY93Z4mPkdwWJ72qiHn9fL/OBtTnM40CdE81Wavu0jWwBkYHh
        vP6UswJp7f5y/ptqpL17Wg8ccc//TBnEGOH27AF5gbwIfypwZbOEuJDTGR8r+gId
        AIAcDTTFjZP+mXF3EB+AU1pHOM68vziambNjces=
        """)

        params = self.session.create_domain_parameters(
            KeyType.DH, decode_dh_domain_parameters(PARAMS), local=True
        )
        self.assertIsInstance(params, DomainParameters)
        self.assertEqual(params[Attribute.PRIME][:4], b"\xad\x10\x7e\x1e")

    @requires(Mechanism.DH_PKCS_PARAMETER_GEN, Mechanism.DH_PKCS_KEY_PAIR_GEN)
    def test_generate_params(self):
        params = self.session.generate_domain_parameters(KeyType.DH, 1024)
        self.assertIsInstance(params, DomainParameters)
        self.assertEqual(params[Attribute.PRIME_BITS], 1024)
        self.assertEqual(len(params[Attribute.PRIME]) * 8, 1024)
        encode_dh_domain_parameters(params)

        # Test encoding the public key
        public, _ = params.generate_keypair()
        encode_dh_public_key(public)



"""
Key handling utilities for DSA keys, domain parameters and signatures..
"""

from asn1crypto.algos import DSASignature
from asn1crypto.core import Integer
from asn1crypto.keys import DSAParams

from pkcs11.constants import Attribute
from pkcs11.util import biginteger


def decode_dsa_domain_parameters(der):
    """
    Decode RFC 3279 DER-encoded Dss-Params.

    :param bytes der: DER-encoded parameters
    :rtype: dict(Attribute,*)
    """

    params = DSAParams.load(der)

    return {
        Attribute.BASE: biginteger(params["g"]),
        Attribute.PRIME: biginteger(params["p"]),
        Attribute.SUBPRIME: biginteger(params["q"]),
    }


def encode_dsa_domain_parameters(obj):
    """
    Encode RFC 3279 DER-encoded Dss-Params.

    :param DomainParameters obj: domain parameters
    :rtype: bytes
    """
    asn1 = DSAParams(
        {
            "g": int.from_bytes(obj[Attribute.BASE], byteorder="big"),
            "p": int.from_bytes(obj[Attribute.PRIME], byteorder="big"),
            "q": int.from_bytes(obj[Attribute.SUBPRIME], byteorder="big"),
        }
    )

    return asn1.dump()


def encode_dsa_public_key(key):
    """
    Encode DSA public key into RFC 3279 DER-encoded format.

    :param PublicKey key: public key
    :rtype: bytes
    """

    asn1 = Integer(int.from_bytes(key[Attribute.VALUE], byteorder="big"))

    return asn1.dump()


def decode_dsa_public_key(der):
    """
    Decode a DSA public key from RFC 3279 DER-encoded format.

    Returns a `biginteger` encoded as bytes.

    :param bytes der: DER-encoded public key
    :rtype: bytes
    """

    asn1 = Integer.load(der)

    return biginteger(asn1)


def encode_dsa_signature(signature):
    """
    Encode a signature (generated by :meth:`pkcs11.SignMixin.sign`) into
    DER-encoded ASN.1 (Dss_Sig_Value) format.

    :param bytes signature: signature as bytes
    :rtype: bytes
    """

    asn1 = DSASignature.from_p1363(signature)

    return asn1.dump()


def decode_dsa_signature(der):
    """
    Decode a DER-encoded ASN.1 (Dss_Sig_Value) signature (as generated by
    OpenSSL/X.509) into PKCS #11 format.

    :param bytes der: DER-encoded signature
    :rtype bytes:
    """

    asn1 = DSASignature.load(der)

    return asn1.to_p1363()



"""
PKCS#11 Sessions
"""

import pkcs11
from pkcs11 import (
    Attribute,
    AttributeSensitive,
    AttributeTypeInvalid,
    ObjectClass,
    PKCS11Error,
)
from pkcs11.attributes import AttributeMapper, handle_bool, handle_str
from pkcs11.exceptions import PinIncorrect, PinLenRange

from . import TOKEN_PIN, TOKEN_SO_PIN, Not, Only, TestCase, requires


class SessionTests(TestCase):
    with_session = False

    @Not.nfast  # Login is required
    @Not.opencryptoki
    def test_open_session(self):
        with self.token.open() as session:
            self.assertIsInstance(session, pkcs11.Session)

    def test_open_session_and_login_user(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            self.assertIsInstance(session, pkcs11.Session)

    @Only.softhsm2  # We don't have credentials to do this for other platforms
    def test_open_session_and_login_so(self):
        with self.token.open(rw=True, so_pin=TOKEN_SO_PIN) as session:
            self.assertIsInstance(session, pkcs11.Session)

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_generate_key(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128)
            self.assertIsInstance(key, pkcs11.Object)
            self.assertIsInstance(key, pkcs11.SecretKey)
            self.assertIsInstance(key, pkcs11.EncryptMixin)

            self.assertIs(key.object_class, pkcs11.ObjectClass.SECRET_KEY)

            # Test GetAttribute
            self.assertIs(key[pkcs11.Attribute.CLASS], pkcs11.ObjectClass.SECRET_KEY)
            self.assertEqual(key[pkcs11.Attribute.TOKEN], False)
            self.assertEqual(key[pkcs11.Attribute.LOCAL], True)
            self.assertEqual(key[pkcs11.Attribute.MODIFIABLE], True)
            self.assertEqual(key[pkcs11.Attribute.LABEL], "")

            # Test SetAttribute
            key[pkcs11.Attribute.LABEL] = "DEMO"

            self.assertEqual(key[pkcs11.Attribute.LABEL], "DEMO")

            # Create another key with no capabilities
            key = session.generate_key(
                pkcs11.KeyType.AES, 128, label="MY KEY", id=b"\1\2\3\4", capabilities=0
            )
            self.assertIsInstance(key, pkcs11.Object)
            self.assertIsInstance(key, pkcs11.SecretKey)
            self.assertNotIsInstance(key, pkcs11.EncryptMixin)

            self.assertEqual(key.label, "MY KEY")

    @requires(pkcs11.Mechanism.RSA_PKCS_KEY_PAIR_GEN, pkcs11.Mechanism.RSA_PKCS)
    def test_generate_keypair(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            pub, priv = session.generate_keypair(pkcs11.KeyType.RSA, 1024)
            self.assertIsInstance(pub, pkcs11.PublicKey)
            self.assertIsInstance(priv, pkcs11.PrivateKey)

            data = b"HELLO WORLD"
            crypttext = pub.encrypt(data, mechanism=pkcs11.Mechanism.RSA_PKCS)
            self.assertNotEqual(data, crypttext)
            text = priv.decrypt(crypttext, mechanism=pkcs11.Mechanism.RSA_PKCS)
            self.assertEqual(data, text)

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_get_objects(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128, label="SAMPLE KEY")

            search = list(
                session.get_objects(
                    {
                        pkcs11.Attribute.LABEL: "SAMPLE KEY",
                    }
                )
            )

            self.assertEqual(len(search), 1)
            self.assertEqual(key, search[0])

    def test_create_object(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key = session.create_object(
                {
                    pkcs11.Attribute.CLASS: pkcs11.ObjectClass.SECRET_KEY,
                    pkcs11.Attribute.KEY_TYPE: pkcs11.KeyType.AES,
                    pkcs11.Attribute.VALUE: b"1" * 16,
                }
            )

            self.assertIsInstance(key, pkcs11.SecretKey)
            self.assertEqual(key.key_length, 128)

    @Not.nfast  # nFast won't destroy objects
    def test_destroy_object(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128, label="SAMPLE KEY")
            key.destroy()

            self.assertEqual(list(session.get_objects()), [])

    @Only.softhsm2
    def test_copy_object(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128, label="SAMPLE KEY")
            new = key.copy(
                {
                    pkcs11.Attribute.LABEL: "SOMETHING ELSE",
                }
            )

            self.assertEqual(set(session.get_objects()), {key, new})

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_get_key(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            session.generate_key(pkcs11.KeyType.AES, 128, label="SAMPLE KEY")

            key = session.get_key(
                label="SAMPLE KEY",
            )
            self.assertIsInstance(key, pkcs11.SecretKey)
            key.encrypt(b"test", mechanism_param=b"IV" * 8)

    def test_get_key_not_found(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            with self.assertRaises(pkcs11.NoSuchKey):
                session.get_key(label="SAMPLE KEY")

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_get_key_vague(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            session.generate_key(pkcs11.KeyType.AES, 128, label="SAMPLE KEY")
            session.generate_key(pkcs11.KeyType.AES, 128, label="SAMPLE KEY 2")

            with self.assertRaises(pkcs11.MultipleObjectsReturned):
                session.get_key(key_type=pkcs11.KeyType.AES)

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_key_search_by_id(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key1 = session.generate_key(pkcs11.KeyType.AES, 128, label="KEY", id=b"1")
            key2 = session.generate_key(pkcs11.KeyType.AES, 128, label="KEY", id=b"2")
            self.assertEqual(session.get_key(id=b"1"), key1)
            self.assertEqual(session.get_key(id=b"2"), key2)
            self.assertNotEqual(session.get_key(id=b"1"), session.get_key(id=b"2"))

    @Not.nfast  # Not supported
    @Not.opencryptoki  # Not supported
    def test_seed_random(self):
        with self.token.open() as session:
            session.seed_random(b"12345678")

    def test_generate_random(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            random = session.generate_random(16 * 8)
            self.assertEqual(len(random), 16)
            # Ensure we didn't get 16 bytes of zeros
            self.assertTrue(all(c != "\0" for c in random))

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_attribute_reading_failures(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128, label="SAMPLE KEY")

            with self.assertRaises(AttributeSensitive):
                key.__getitem__(Attribute.VALUE)

            with self.assertRaises(AttributeTypeInvalid):
                key.__getitem__(Attribute.CERTIFICATE_TYPE)

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_bulk_attribute_raise_error_if_no_result(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128, label="SAMPLE KEY")

            with self.assertRaises(AttributeSensitive):
                key.get_attributes([Attribute.VALUE])

            with self.assertRaises(AttributeTypeInvalid):
                key.get_attributes([Attribute.CERTIFICATE_TYPE])
            # we can't know which error code the token will choose here
            with self.assertRaises(PKCS11Error):
                key.get_attributes([Attribute.VALUE, Attribute.CERTIFICATE_TYPE])

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_bulk_attribute_partial_success_sensitive_attribute(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128, label="SAMPLE KEY")
            result = key.get_attributes([Attribute.LABEL, Attribute.VALUE, Attribute.CLASS])
            expected = {Attribute.LABEL: "SAMPLE KEY", Attribute.CLASS: ObjectClass.SECRET_KEY}
            self.assertDictEqual(expected, result)

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_bulk_attribute_partial_success_irrelevant_attribute(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128, label="SAMPLE KEY", id=b"a")

            result = key.get_attributes(
                [Attribute.LABEL, Attribute.CERTIFICATE_TYPE, Attribute.CLASS, Attribute.ID]
            )
            expected = {
                Attribute.LABEL: "SAMPLE KEY",
                Attribute.CLASS: ObjectClass.SECRET_KEY,
                Attribute.ID: b"a",
            }
            self.assertDictEqual(expected, result)

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_bulk_attribute_partial_success_with_some_empty_attrs(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128, label="", id=b"")

            result = key.get_attributes(
                [Attribute.LABEL, Attribute.CLASS, Attribute.VALUE, Attribute.ID]
            )
            expected = {
                Attribute.LABEL: "",
                Attribute.CLASS: ObjectClass.SECRET_KEY,
                Attribute.ID: b"",
            }
            self.assertDictEqual(expected, result)

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_bulk_attribute_only_empty_attrs(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128, label="", id=b"")

            result = key.get_attributes([Attribute.LABEL, Attribute.ID])
            expected = {
                Attribute.LABEL: "",
                Attribute.ID: b"",
            }
            self.assertDictEqual(expected, result)

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_bulk_attribute_empty_key_list(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128, label="SAMPLE KEY")

            result = key.get_attributes([])
            self.assertDictEqual({}, result)

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_custom_attribute_mapper(self):
        custom_mapper = AttributeMapper()
        custom_mapper.register_handler(Attribute.ID, *handle_str)

        with self.token.open(user_pin=TOKEN_PIN, attribute_mapper=custom_mapper) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128, id="SAMPLE KEY")
            id_attr = key[Attribute.ID]
            self.assertIsInstance(id_attr, str)
            self.assertEqual("SAMPLE KEY", id_attr)

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_set_unsupported_attribute(self):
        with self.token.open(user_pin=TOKEN_PIN) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128, label="TEST")

            with self.assertRaises(NotImplementedError):
                key[0xDEADBEEF] = b"1234"

    @requires(pkcs11.Mechanism.AES_KEY_GEN)
    def test_treat_empty_bool_as_false(self):
        class CustomMapper(AttributeMapper):
            # contrived handler that decodes the 'ID' attribute as a bool
            def _handler(self, key):
                orig = super()._handler(key)
                if key == Attribute.ID:
                    return orig[0], handle_bool[1]
                return orig

        with self.token.open(user_pin=TOKEN_PIN, attribute_mapper=CustomMapper()) as session:
            key = session.generate_key(pkcs11.KeyType.AES, 128, id=b"")
            bool_read = key[Attribute.ID]
            self.assertIsInstance(bool_read, bool)
            self.assertFalse(bool_read, False)

    @Only.softhsm2
    def test_set_pin(self):
        old_token_pin = TOKEN_PIN
        new_token_pin = f"{TOKEN_PIN}56"

        with self.token.open(rw=True, user_pin=old_token_pin) as session:
            session.set_pin(old_token_pin, new_token_pin)

        with self.token.open(user_pin=new_token_pin) as session:
            self.assertIsInstance(session, pkcs11.Session)

        with self.token.open(rw=True, user_pin=new_token_pin) as session:
            session.set_pin(new_token_pin, old_token_pin)

        with self.token.open(user_pin=old_token_pin) as session:
            self.assertIsInstance(session, pkcs11.Session)

        with self.token.open(rw=True, user_pin=old_token_pin) as session:
            with self.assertRaises(AttributeError):
                session.set_pin(None, new_token_pin)
            with self.assertRaises(AttributeError):
                session.set_pin(old_token_pin, None)
            with self.assertRaises(PinLenRange):
                session.set_pin(old_token_pin, "")
            with self.assertRaises(PinIncorrect):
                session.set_pin("", new_token_pin)

    @Only.softhsm2
    def test_init_pin(self):
        new_token_pin = f"{TOKEN_PIN}56"

        with self.token.open(rw=True, so_pin=TOKEN_SO_PIN) as session:
            session.init_pin(new_token_pin)

        with self.token.open(rw=True, user_pin=new_token_pin) as session:
            self.assertIsInstance(session, pkcs11.Session)

        with self.token.open(rw=True, so_pin=TOKEN_SO_PIN) as session:
            session.init_pin(TOKEN_PIN)

        with self.token.open(rw=True, user_pin=TOKEN_PIN) as session:
            self.assertIsInstance(session, pkcs11.Session)

        with self.token.open(rw=True, so_pin=TOKEN_SO_PIN) as session:
            with self.assertRaises(AttributeError):
                session.init_pin(None)
            with self.assertRaises(PinLenRange):
                session.init_pin("")



"""
PKCS#11 AES Secret Keys
"""

from parameterized import parameterized

import pkcs11
from pkcs11 import ArgumentsBad, CTRParams, GCMParams, Mechanism, PKCS11Error

from . import FIXME, TestCase, requires


class AESTests(TestCase):
    @requires(Mechanism.AES_KEY_GEN)
    def setUp(self):
        super().setUp()
        self.key = self.session.generate_key(pkcs11.KeyType.AES, 128)

    @requires(Mechanism.AES_CBC_PAD)
    def test_encrypt(self):
        data = b"INPUT DATA"
        iv = b"0" * 16

        crypttext = self.key.encrypt(data, mechanism_param=iv)
        self.assertIsInstance(crypttext, bytes)
        self.assertNotEqual(data, crypttext)
        # We should be aligned to the block size
        self.assertEqual(len(crypttext), 16)
        # Ensure we didn't just get 16 nulls
        self.assertFalse(all(c == "\0" for c in crypttext))

        text = self.key.decrypt(crypttext, mechanism_param=iv)
        self.assertEqual(data, text)

    @requires(Mechanism.AES_CBC_PAD)
    def test_encrypt_undersized_buffer(self):
        data = b"INPUT DATA" * 200
        iv = b"0" * 16

        crypttext = self.key.encrypt(data, mechanism_param=iv, buffer_size=32)
        text = self.key.decrypt(crypttext, mechanism_param=iv, buffer_size=32)
        self.assertEqual(data, text)

    @requires(Mechanism.AES_CBC_PAD)
    def test_encrypt_stream(self):
        data = (
            b"I" * 16,
            b"N" * 16,
            b"P" * 16,
            b"U" * 16,
            b"T" * 10,  # don't align to the blocksize
        )
        iv = b"0" * 16

        cryptblocks = list(self.key.encrypt(data, mechanism_param=iv))

        self.assertEqual(len(cryptblocks), len(data) + 1)

        crypttext = b"".join(cryptblocks)

        self.assertNotEqual(b"".join(data), crypttext)
        # We should be aligned to the block size
        self.assertEqual(len(crypttext) % 16, 0)
        # Ensure we didn't just get 16 nulls
        self.assertFalse(all(c == "\0" for c in crypttext))

        text = b"".join(self.key.decrypt(cryptblocks, mechanism_param=iv))
        self.assertEqual(b"".join(data), text)

    @requires(Mechanism.AES_CBC_PAD)
    def test_encrypt_stream_interrupt_releases_operation(self):
        data = (
            b"I" * 16,
            b"N" * 16,
            b"P" * 16,
            b"U" * 16,
            b"T" * 10,
        )
        iv = b"0" * 16

        def _data_with_error():
            yield data[0]
            yield data[1]
            yield data[2]
            raise ValueError

        def attempt_encrypt():
            list(self.key.encrypt(_data_with_error(), mechanism_param=iv))

        self.assertRaises(ValueError, attempt_encrypt)

        cryptblocks = list(self.key.encrypt(data, mechanism_param=iv))
        text = b"".join(self.key.decrypt(cryptblocks, mechanism_param=iv))
        self.assertEqual(b"".join(data), text)

    @requires(Mechanism.AES_CBC_PAD)
    def test_encrypt_stream_gc_releases_operation(self):
        data = (
            b"I" * 16,
            b"N" * 16,
            b"P" * 16,
            b"U" * 16,
            b"T" * 10,
        )
        iv = b"0" * 16

        attempt = self.key.encrypt(data, mechanism_param=iv)
        next(attempt)
        del attempt

        cryptblocks = list(self.key.encrypt(data, mechanism_param=iv))
        text = b"".join(self.key.decrypt(cryptblocks, mechanism_param=iv))
        self.assertEqual(b"".join(data), text)

    @requires(Mechanism.AES_CBC_PAD)
    def test_encrypt_stream_undersized_buffers(self):
        data = (
            b"I" * 3189,
            b"N" * 3284,
            b"P" * 5812,
            b"U" * 2139,
            b"T" * 5851,
        )
        iv = b"0" * 16

        cryptblocks = list(self.key.encrypt(data, mechanism_param=iv, buffer_size=256))

        self.assertEqual(len(cryptblocks), len(data) + 1)

        crypttext = b"".join(cryptblocks)

        self.assertNotEqual(b"".join(data), crypttext)
        text = b"".join(self.key.decrypt(cryptblocks, mechanism_param=iv, buffer_size=5))
        self.assertEqual(b"".join(data), text)

    @requires(Mechanism.AES_CBC_PAD)
    def test_encrypt_whacky_sizes(self):
        data = [(char * ord(char)).encode("utf-8") for char in "HELLO WORLD"]
        iv = b"0" * 16

        cryptblocks = list(self.key.encrypt(data, mechanism_param=iv))
        textblocks = list(self.key.decrypt(cryptblocks, mechanism_param=iv))

        self.assertEqual(b"".join(data), b"".join(textblocks))

    @requires(Mechanism.AES_CBC_PAD)
    def test_encrypt_big_string(self):
        data = b"HELLO WORLD" * 1024

        iv = self.session.generate_random(128)
        crypttext = self.key.encrypt(data, mechanism_param=iv)
        text = self.key.decrypt(crypttext, mechanism_param=iv)

        self.assertEqual(text, data)

    @requires(Mechanism.AES_MAC)
    def test_sign(self):
        data = b"HELLO WORLD"

        signature = self.key.sign(data)
        self.assertIsNotNone(signature)
        self.assertIsInstance(signature, bytes)
        self.assertTrue(self.key.verify(data, signature))
        self.assertFalse(self.key.verify(data, b"1234"))

    @requires(Mechanism.AES_MAC)
    def test_sign_stream(self):
        data = (
            b"I" * 16,
            b"N" * 16,
            b"P" * 16,
            b"U" * 16,
            b"T" * 10,  # don't align to the blocksize
        )

        signature = self.key.sign(data)
        self.assertIsNotNone(signature)
        self.assertIsInstance(signature, bytes)
        self.assertTrue(self.key.verify(data, signature))

    @requires(Mechanism.AES_KEY_WRAP)
    @FIXME.opencryptoki  # can't set key attributes
    def test_wrap(self):
        key = self.session.generate_key(
            pkcs11.KeyType.AES,
            128,
            template={
                pkcs11.Attribute.EXTRACTABLE: True,
                pkcs11.Attribute.SENSITIVE: False,
            },
        )
        data = self.key.wrap_key(key)

        key2 = self.key.unwrap_key(
            pkcs11.ObjectClass.SECRET_KEY,
            pkcs11.KeyType.AES,
            data,
            template={
                pkcs11.Attribute.EXTRACTABLE: True,
                pkcs11.Attribute.SENSITIVE: False,
            },
        )

        self.assertEqual(key[pkcs11.Attribute.VALUE], key2[pkcs11.Attribute.VALUE])

    @parameterized.expand(
        [
            ("POSITIVE_128_BIT", 128, 16, TestCase.assertIsNotNone),
            ("POSITIVE_128_BIT_LONG_IV", 128, 32, TestCase.assertIsNotNone),
            ("NEGATIVE_128_BIT_BAD_IV", 128, 15, TestCase.assertIsNone),
            ("POSITIVE_256_BIT_LONG_IV", 256, 32, TestCase.assertIsNotNone),
            ("NEGATIVE_256_BIT_SHORT_IV", 256, 16, TestCase.assertIsNone),
            ("NEGATIVE_256_BIT_BAD_IV", 256, 31, TestCase.assertIsNone),
        ]
    )
    @requires(Mechanism.AES_ECB_ENCRYPT_DATA)
    @FIXME.opencryptoki  # can't set key attributes
    def test_derive_using_ecb_encrypt(self, test_type, test_key_length, iv_length, assert_fn):
        """Function to test AES Key Derivation using the ECB_ENCRYPT Mechanism.

        Refer to Section 2.15 of http://docs.oasis-open.org/pkcs11/pkcs11-curr/v2.40/errata01/os/pkcs11-curr-v2.40-errata01-os-complete.html#_Toc441850521
        """

        # Create the Master Key
        capabilities = pkcs11.defaults.DEFAULT_KEY_CAPABILITIES[pkcs11.KeyType.AES]
        capabilities |= pkcs11.MechanismFlag.DERIVE
        key = self.session.generate_key(
            pkcs11.KeyType.AES,
            key_length=test_key_length,
            capabilities=capabilities,
            template={
                pkcs11.Attribute.EXTRACTABLE: True,
                pkcs11.Attribute.DERIVE: True,
                pkcs11.Attribute.SENSITIVE: False,
            },
        )

        self.assertTrue(
            key is not None, "Failed to create {}-bit Master Key".format(test_key_length)
        )

        # Derive a Key from the Master Key
        iv = b"0" * iv_length
        try:
            derived_key = key.derive_key(
                pkcs11.KeyType.AES,
                key_length=test_key_length,
                capabilities=capabilities,
                mechanism=Mechanism.AES_ECB_ENCRYPT_DATA,
                mechanism_param=iv,
                template={
                    pkcs11.Attribute.EXTRACTABLE: True,
                    pkcs11.Attribute.SENSITIVE: False,
                },
            )
        except (pkcs11.exceptions.MechanismParamInvalid, pkcs11.exceptions.FunctionFailed):
            derived_key = None

        assert_fn(self, derived_key, "{}-bit Key Derivation Failure".format(test_key_length))

    @parameterized.expand(
        [
            ("POSITIVE_128_BIT", 128, 16),
            ("POSITIVE_256_BIT_LONG_IV", 256, 32),
        ]
    )
    @requires(Mechanism.AES_ECB_ENCRYPT_DATA)
    @FIXME.opencryptoki  # can't set key attributes
    def test_encrypt_with_key_derived_using_ecb_encrypt(
        self, test_type, test_key_length, iv_length
    ):
        """Function to test Data Encryption/Decryption using a Derived AES Key.

        Function to test Data Encryption/Decryption using an AES Key
        Derived by the ECB_ENCRYPT Mechanism.

        Refer to Section 2.15 of http://docs.oasis-open.org/pkcs11/pkcs11-curr/v2.40/errata01/os/pkcs11-curr-v2.40-errata01-os-complete.html#_Toc441850521
        """

        # Create the Master Key
        capabilities = pkcs11.defaults.DEFAULT_KEY_CAPABILITIES[pkcs11.KeyType.AES]
        capabilities |= pkcs11.MechanismFlag.DERIVE
        key = self.session.generate_key(
            pkcs11.KeyType.AES,
            key_length=test_key_length,
            capabilities=capabilities,
            template={
                pkcs11.Attribute.EXTRACTABLE: True,
                pkcs11.Attribute.DERIVE: True,
                pkcs11.Attribute.SENSITIVE: False,
            },
        )

        self.assertTrue(
            key is not None, "Failed to create {}-bit Master Key".format(test_key_length)
        )

        # Derive a Key from the Master Key
        iv = b"0" * iv_length
        try:
            derived_key = key.derive_key(
                pkcs11.KeyType.AES,
                key_length=test_key_length,
                capabilities=capabilities,
                mechanism=Mechanism.AES_ECB_ENCRYPT_DATA,
                mechanism_param=iv,
                template={
                    pkcs11.Attribute.EXTRACTABLE: True,
                    pkcs11.Attribute.SENSITIVE: False,
                },
            )
        except (pkcs11.exceptions.MechanismParamInvalid, pkcs11.exceptions.FunctionFailed):
            derived_key = None

        self.assertTrue(
            derived_key is not None, "Failed to derive {}-bit Derived Key".format(test_key_length)
        )

        # Test capability of Key to Encrypt/Decrypt data
        data = b"HELLO WORLD" * 1024

        iv = self.session.generate_random(128)
        crypttext = self.key.encrypt(data, mechanism_param=iv)
        text = self.key.decrypt(crypttext, mechanism_param=iv)

        self.assertEqual(text, data)

    @parameterized.expand(
        [
            ("POSITIVE_128_BIT", 128, 16, 16, TestCase.assertIsNotNone),
            ("POSITIVE_128_BIT_LONG_DATA", 128, 16, 64, TestCase.assertIsNotNone),
            ("NEGATIVE_128_BIT_BAD_IV", 128, 15, 16, TestCase.assertIsNone),
            ("NEGATIVE_128_BIT_BAD_DATA", 128, 16, 31, TestCase.assertIsNone),
            ("POSITIVE_256_BIT", 256, 16, 32, TestCase.assertIsNotNone),
            ("POSITIVE_256_BIT_LONG_DATA", 256, 16, 64, TestCase.assertIsNotNone),
            ("NEGATIVE_256_BIT_BAD_IV", 256, 15, 16, TestCase.assertIsNone),
            ("NEGATIVE_256_BIT_BAD_DATA", 256, 16, 31, TestCase.assertIsNone),
            ("NEGATIVE_256_BIT_SHORT_DATA", 256, 16, 16, TestCase.assertIsNone),
        ]
    )
    @requires(Mechanism.AES_CBC_ENCRYPT_DATA)
    @FIXME.opencryptoki  # can't set key attributes
    def test_derive_using_cbc_encrypt(
        self, test_type, test_key_length, iv_length, data_length, assert_fn
    ):
        """Function to test AES Key Derivation using the CBC_ENCRYPT Mechanism.

        Refer to Section 2.15 of http://docs.oasis-open.org/pkcs11/pkcs11-curr/v2.40/errata01/os/pkcs11-curr-v2.40-errata01-os-complete.html#_Toc441850521
        """

        # Create the Master Key
        capabilities = pkcs11.defaults.DEFAULT_KEY_CAPABILITIES[pkcs11.KeyType.AES]
        capabilities |= pkcs11.MechanismFlag.DERIVE
        key = self.session.generate_key(
            pkcs11.KeyType.AES,
            key_length=test_key_length,
            capabilities=capabilities,
            template={
                pkcs11.Attribute.EXTRACTABLE: True,
                pkcs11.Attribute.DERIVE: True,
                pkcs11.Attribute.SENSITIVE: False,
            },
        )

        self.assertTrue(
            key is not None, "Failed to create {}-bit Master Key".format(test_key_length)
        )

        # Derive a Key from the Master Key
        iv = b"0" * iv_length
        data = b"1" * data_length
        try:
            derived_key = key.derive_key(
                pkcs11.KeyType.AES,
                key_length=test_key_length,
                capabilities=capabilities,
                mechanism=Mechanism.AES_CBC_ENCRYPT_DATA,
                mechanism_param=(iv, data),
                template={
                    pkcs11.Attribute.EXTRACTABLE: True,
                    pkcs11.Attribute.SENSITIVE: False,
                },
            )
        except (
            pkcs11.exceptions.MechanismParamInvalid,
            pkcs11.exceptions.FunctionFailed,
            IndexError,
        ):
            derived_key = None

        assert_fn(self, derived_key, "{}-bit Key Derivation Failure".format(test_key_length))

    @parameterized.expand(
        [
            ("POSITIVE_128_BIT", 128, 16, 16),
            ("POSITIVE_256_BIT", 256, 16, 32),
            ("POSITIVE_256_BIT_LONG_DATA", 256, 16, 64),
        ]
    )
    @requires(Mechanism.AES_CBC_ENCRYPT_DATA)
    @FIXME.opencryptoki  # can't set key attributes
    def test_encrypt_with_key_derived_using_cbc_encrypt(
        self, test_type, test_key_length, iv_length, data_length
    ):
        """Function to test Data Encryption/Decryption using a Derived AES Key.

        Function to test Data Encryption/Decryption using an AES Key
        Derived by the CBC_ENCRYPT Mechanism.

        Refer to Section 2.15 of http://docs.oasis-open.org/pkcs11/pkcs11-curr/v2.40/errata01/os/pkcs11-curr-v2.40-errata01-os-complete.html#_Toc441850521
        """

        # Create the Master Key
        capabilities = pkcs11.defaults.DEFAULT_KEY_CAPABILITIES[pkcs11.KeyType.AES]
        capabilities |= pkcs11.MechanismFlag.DERIVE
        key = self.session.generate_key(
            pkcs11.KeyType.AES,
            key_length=test_key_length,
            capabilities=capabilities,
            template={
                pkcs11.Attribute.EXTRACTABLE: True,
                pkcs11.Attribute.DERIVE: True,
                pkcs11.Attribute.SENSITIVE: False,
            },
        )

        self.assertTrue(
            key is not None, "Failed to create {}-bit Master Key".format(test_key_length)
        )

        # Derive a Key from the Master Key
        iv = b"0" * iv_length
        data = b"1" * data_length
        try:
            derived_key = key.derive_key(
                pkcs11.KeyType.AES,
                key_length=test_key_length,
                capabilities=capabilities,
                mechanism=Mechanism.AES_CBC_ENCRYPT_DATA,
                mechanism_param=(iv, data),
                template={
                    pkcs11.Attribute.EXTRACTABLE: True,
                    pkcs11.Attribute.SENSITIVE: False,
                },
            )
        except (
            pkcs11.exceptions.MechanismParamInvalid,
            pkcs11.exceptions.FunctionFailed,
            IndexError,
        ):
            derived_key = None

        self.assertTrue(
            derived_key is not None, "Failed to derive {}-bit Derived Key".format(test_key_length)
        )

        # Test capability of Key to Encrypt/Decrypt data
        data = b"HELLO WORLD" * 1024

        iv = self.session.generate_random(128)
        crypttext = self.key.encrypt(data, mechanism_param=iv)
        text = self.key.decrypt(crypttext, mechanism_param=iv)

        self.assertEqual(text, data)

    @requires(Mechanism.AES_GCM)
    def test_encrypt_gcm(self):
        data = b"INPUT DATA"
        nonce = b"0" * 12

        crypttext = self.key.encrypt(
            data, mechanism=Mechanism.AES_GCM, mechanism_param=GCMParams(nonce)
        )
        self.assertIsInstance(crypttext, bytes)
        self.assertNotEqual(data, crypttext)
        text = self.key.decrypt(
            crypttext, mechanism=Mechanism.AES_GCM, mechanism_param=GCMParams(nonce)
        )
        self.assertEqual(data, text)

    def test_gcm_nonce_size_limit(self):
        def _inst():
            return GCMParams(nonce=b"0" * 13)

        self.assertRaises(ArgumentsBad, _inst)

    @requires(Mechanism.AES_GCM)
    def test_encrypt_gcm_with_aad(self):
        data = b"INPUT DATA"
        nonce = b"0" * 12

        crypttext = self.key.encrypt(
            data, mechanism=Mechanism.AES_GCM, mechanism_param=GCMParams(nonce, b"foo")
        )
        self.assertIsInstance(crypttext, bytes)
        self.assertNotEqual(data, crypttext)
        text = self.key.decrypt(
            crypttext, mechanism=Mechanism.AES_GCM, mechanism_param=GCMParams(nonce, b"foo")
        )
        self.assertEqual(data, text)

    @requires(Mechanism.AES_GCM)
    def test_encrypt_gcm_with_mismatching_nonces(self):
        data = b"INPUT DATA"
        nonce1 = b"0" * 12
        nonce2 = b"1" * 12

        crypttext = self.key.encrypt(
            data, mechanism=Mechanism.AES_GCM, mechanism_param=GCMParams(nonce1, b"foo")
        )
        self.assertIsInstance(crypttext, bytes)
        self.assertNotEqual(data, crypttext)
        # This should be EncryptedDataInvalid, but in practice not all tokens support this
        with self.assertRaises(PKCS11Error):
            self.key.decrypt(
                crypttext, mechanism=Mechanism.AES_GCM, mechanism_param=GCMParams(nonce2, b"foo")
            )

    @requires(Mechanism.AES_GCM)
    def test_encrypt_gcm_with_mismatching_aad(self):
        data = b"INPUT DATA"
        nonce = b"0" * 12

        crypttext = self.key.encrypt(
            data, mechanism=Mechanism.AES_GCM, mechanism_param=GCMParams(nonce, b"foo")
        )
        self.assertIsInstance(crypttext, bytes)
        self.assertNotEqual(data, crypttext)
        with self.assertRaises(PKCS11Error):
            self.key.decrypt(
                crypttext, mechanism=Mechanism.AES_GCM, mechanism_param=GCMParams(nonce, b"bar")
            )

    @requires(Mechanism.AES_GCM)
    def test_encrypt_gcm_with_custom_tag_length(self):
        data = b"INPUT DATA"
        nonce = b"0" * 12

        crypttext = self.key.encrypt(
            data, mechanism=Mechanism.AES_GCM, mechanism_param=GCMParams(nonce, b"foo", 120)
        )
        self.assertIsInstance(crypttext, bytes)
        self.assertNotEqual(data, crypttext)
        text = self.key.decrypt(
            crypttext, mechanism=Mechanism.AES_GCM, mechanism_param=GCMParams(nonce, b"foo", 120)
        )
        self.assertEqual(data, text)

        # This should be EncryptedDataInvalid, but in practice not all tokens support this
        with self.assertRaises(PKCS11Error):
            text = self.key.decrypt(
                crypttext,
                mechanism=Mechanism.AES_GCM,
                mechanism_param=GCMParams(nonce, b"foo", 128),
            )

    @parameterized.expand(
        [
            (b""),
            (b"0" * 12),
            (b"0" * 15),
        ]
    )
    @requires(Mechanism.AES_CTR)
    @FIXME.opencryptoki  # opencryptoki incorrectly forces AES-CTR input to be padded
    def test_encrypt_ctr(self, nonce):
        data = b"INPUT DATA SEVERAL BLOCKS LONG SO THE COUNTER GOES UP A FEW TIMES" * 20

        crypttext = self.key.encrypt(
            data, mechanism=Mechanism.AES_CTR, mechanism_param=CTRParams(nonce)
        )
        self.assertIsInstance(crypttext, bytes)
        self.assertNotEqual(data, crypttext)
        text = self.key.decrypt(
            crypttext, mechanism=Mechanism.AES_CTR, mechanism_param=CTRParams(nonce)
        )
        self.assertEqual(data, text)

    @requires(Mechanism.AES_CTR)
    def test_encrypt_ctr_exactly_padded(self):
        # let's still verify the "restricted" AES-CTR supported by opencryptoki
        data = b"PADDED INPUT DATA TO MAKE OPENCRYPTOKI HAPPY" * 16
        nonce = b"0" * 15

        crypttext = self.key.encrypt(
            data, mechanism=Mechanism.AES_CTR, mechanism_param=CTRParams(nonce)
        )
        self.assertIsInstance(crypttext, bytes)
        self.assertNotEqual(data, crypttext)
        text = self.key.decrypt(
            crypttext, mechanism=Mechanism.AES_CTR, mechanism_param=CTRParams(nonce)
        )
        self.assertEqual(data, text)

    def test_ctr_nonce_size_limit(self):
        def _inst():
            return CTRParams(nonce=b"0" * 16)

        self.assertRaises(ArgumentsBad, _inst)

    @requires(Mechanism.AES_CTR)
    @FIXME.opencryptoki  # opencryptoki incorrectly forces AES-CTR input to be padded
    def test_encrypt_ctr_nonce_mismatch(self):
        data = b"INPUT DATA SEVERAL BLOCKS LONG SO THE COUNTER GOES UP A FEW TIMES" * 20
        crypttext = self.key.encrypt(
            data, mechanism=Mechanism.AES_CTR, mechanism_param=CTRParams(b"0" * 12)
        )
        self.assertIsInstance(crypttext, bytes)
        self.assertNotEqual(data, crypttext)
        text = self.key.decrypt(
            crypttext, mechanism=Mechanism.AES_CTR, mechanism_param=CTRParams(b"1" * 12)
        )
        self.assertNotEqual(data, text)

    @parameterized.expand(
        [
            (
                "ae6852f8121067cc4bf7a5765577f39e",
                b"Single block msg",
                "00000030",
                "0000000000000000",
                "e4095d4fb7a7b3792d6175a3261311b8",
            ),
            (
                "7e24067817fae0d743d6ce1f32539163",
                bytes(range(0x20)),
                "006cb6db",
                "c0543b59da48d90b",
                "5104a106168a72d9790d41ee8edad388eb2e1efc46da57c8fce630df9141be28",
            ),
            (
                "7691be035e5020a8ac6e618529f9a0dc",
                bytes(range(0x24)),
                "00e0017b",
                "27777f3f4a1786f0",
                "c1cf48a89f2ffdd9cf4652e9efdb72d74540a42bde6d7836d59a5ceaaef3105325b2072f",
            ),
            (
                "16af5b145fc9f579c175f93e3bfb0eed863d06ccfdb78515",
                b"Single block msg",
                "00000048",
                "36733c147d6d93cb",
                "4b55384fe259c9c84e7935a003cbe928",
            ),
            (
                "7c5cb2401b3dc33c19e7340819e0f69c678c3db8e6f6a91a",
                bytes(range(0x20)),
                "0096b03b",
                "020c6eadc2cb500d",
                "453243fc609b23327edfaafa7131cd9f8490701c5ad4a79cfc1fe0ff42f4fb00",
            ),
            (
                "02bf391ee8ecb159b959617b0965279bf59b60a786d3e0fe",
                bytes(range(0x24)),
                "0007bdfd",
                "5cbd60278dcc0912",
                "96893fc55e5c722f540b7dd1ddf7e758d288bc95c69165884536c811662f2188abee0935",
            ),
            (
                "776beff2851db06f4c8a0542c8696f6c6a81af1eec96b4d37fc1d689e6c1c104",
                b"Single block msg",
                "00000060",
                "db5672c97aa8f0b2",
                "145ad01dbf824ec7560863dc71e3e0c0",
            ),
            (
                "f6d66d6bd52d59bb0796365879eff886c66dd51a5b6a99744b50590c87a23884",
                bytes(range(0x20)),
                "00faac24",
                "c1585ef15a43d875",
                "f05e231b3894612c49ee000b804eb2a9b8306b508f839d6a5530831d9344af1c",
            ),
        ]
    )
    # https://github.com/opencryptoki/opencryptoki/issues/881
    @FIXME.opencryptoki
    @requires(Mechanism.AES_CTR)
    def test_aes_ctr_test_vector(self, key, plaintext, nonce, iv, expected_ciphertext):
        """Official test vectors from RFC 3686"""
        key = self.session.create_object(
            {
                pkcs11.Attribute.CLASS: pkcs11.ObjectClass.SECRET_KEY,
                pkcs11.Attribute.KEY_TYPE: pkcs11.KeyType.AES,
                pkcs11.Attribute.VALUE: bytes.fromhex(key),
            }
        )

        params = CTRParams(bytes.fromhex(nonce) + bytes.fromhex(iv))
        ciphertext = key.encrypt(plaintext, mechanism_param=params, mechanism=Mechanism.AES_CTR)
        self.assertEqual(bytes.fromhex(expected_ciphertext), ciphertext)

    @parameterized.expand(
        [
            (
                "00000000000000000000000000000000",
                "",
                "",
                "000000000000000000000000",
                "",
                "58e2fccefa7e3061367f1d57a4e7455a",
            ),
            (
                "00000000000000000000000000000000",
                "00000000000000000000000000000000",
                "",
                "000000000000000000000000",
                "0388dace60b6a392f328c2b971b2fe78",
                "ab6e47d42cec13bdf53a67b21257bddf",
            ),
            (
                "feffe9928665731c6d6a8f9467308308",
                "d9313225f88406e5a55909c5aff5269a86a7a9531534f7da2e4c303d8a318a721c3c0c95956809532fcf0e2449a6b525b16aedf5aa0de657ba637b391aafd255",
                "",
                "cafebabefacedbaddecaf888",
                "42831ec2217774244b7221b784d0d49ce3aa212f2c02a4e035c17e2329aca12e21d514b25466931c7d8f6a5aac84aa051ba30b396a0aac973d58e091473f5985",
                "4d5c2af327cd64a62cf35abd2ba6fab4",
            ),
            (
                "feffe9928665731c6d6a8f9467308308",
                "d9313225f88406e5a55909c5aff5269a86a7a9531534f7da2e4c303d8a318a721c3c0c95956809532fcf0e2449a6b525b16aedf5aa0de657ba637b39",
                "feedfacedeadbeeffeedfacedeadbeefabaddad2",
                "cafebabefacedbaddecaf888",
                "42831ec2217774244b7221b784d0d49ce3aa212f2c02a4e035c17e2329aca12e21d514b25466931c7d8f6a5aac84aa051ba30b396a0aac973d58e091",
                "5bc94fbc3221a5db94fae95ae7121a47",
            ),
        ],
    )
    @requires(Mechanism.AES_GCM)
    def test_aes_gcm_test_vector(
        self, key, plaintext, aad, nonce, expected_ciphertext, expected_tag
    ):
        """Some test vectors from McGrew-Viega"""
        key = self.session.create_object(
            {
                pkcs11.Attribute.CLASS: pkcs11.ObjectClass.SECRET_KEY,
                pkcs11.Attribute.KEY_TYPE: pkcs11.KeyType.AES,
                pkcs11.Attribute.VALUE: bytes.fromhex(key),
            }
        )

        params = GCMParams(nonce=bytes.fromhex(nonce), aad=bytes.fromhex(aad))
        result = key.encrypt(
            bytes.fromhex(plaintext), mechanism_param=params, mechanism=Mechanism.AES_GCM
        )
        expected_output = bytes.fromhex(expected_ciphertext) + bytes.fromhex(expected_tag)
        self.assertEqual(expected_output, result)



"""
PKCS#11 Thread Safety tests

Even though you have a session construct it turns out the expectation of
PKCS#11 is that you have a single session per process.
"""

import threading

import pkcs11

from . import Not, TestCase, requires


@Not.nfast  # Deadlocks nfast ... something wrong with threading?
class ThreadingTests(TestCase):
    @requires(pkcs11.Mechanism.AES_KEY_GEN, pkcs11.Mechanism.AES_CBC_PAD)
    def test_concurrency(self):
        # Multiplexing a session between processes
        self.session.generate_key(pkcs11.KeyType.AES, 128, label="LOOK ME UP")

        test_passed = [True]

        def thread_work():
            try:
                data = b"1234" * 1024 * 1024  # Multichunk files
                iv = self.session.generate_random(128)
                key = self.session.get_key(label="LOOK ME UP")
                self.assertIsNotNone(key.encrypt(data, mechanism_param=iv))
            except pkcs11.PKCS11Error:
                test_passed[0] = False
                raise

        threads = [threading.Thread(target=thread_work) for _ in range(10)]

        for thread in threads:
            thread.start()

        # join each thread
        for thread in threads:
            thread.join()

        self.assertTrue(test_passed[0])



"""
PKCS#11 Elliptic Curve Cryptography.
"""

import base64

from asn1crypto.keys import PrivateKeyAlgorithmId

import pkcs11
from pkcs11 import KDF, Attribute, KeyType, Mechanism
from pkcs11.util.ec import (
    decode_ec_private_key,
    decode_ec_public_key,
    decode_ecdsa_signature,
    encode_ec_public_key,
    encode_named_curve_parameters,
)

from . import Only, TestCase, requires


class ECCTests(TestCase):
    @requires(Mechanism.EC_KEY_PAIR_GEN, Mechanism.ECDSA)
    def test_sign_ecdsa(self):
        parameters = self.session.create_domain_parameters(
            KeyType.EC,
            {Attribute.EC_PARAMS: encode_named_curve_parameters("secp256r1")},
            local=True,
        )

        pub, priv = parameters.generate_keypair()

        mechanism = Mechanism.ECDSA
        data = b"HI BOB!"
        ecdsa = priv.sign(data, mechanism=mechanism)
        self.assertTrue(pub.verify(data, ecdsa, mechanism=mechanism))

    @requires(Mechanism.EC_KEY_PAIR_GEN, Mechanism.ECDH1_DERIVE)
    def test_derive_key(self):
        # DER encoded EC params from OpenSSL
        # openssl ecparam -out ec_param.der -name prime192v1
        ecparams = base64.b64decode(b"BggqhkjOPQMBAQ==")

        parameters = self.session.create_domain_parameters(
            KeyType.EC,
            {
                Attribute.EC_PARAMS: ecparams,
            },
            local=True,
        )
        alice_pub, alice_priv = parameters.generate_keypair()
        alice_value = alice_pub[Attribute.EC_POINT]

        bob_pub, bob_priv = parameters.generate_keypair()
        bob_value = bob_pub[Attribute.EC_POINT]

        self.assertNotEqual(alice_value, bob_value)

        alice_session = alice_priv.derive_key(
            KeyType.AES, 128, mechanism_param=(KDF.NULL, None, bob_value)
        )

        bob_session = bob_priv.derive_key(
            KeyType.AES, 128, mechanism_param=(KDF.NULL, None, alice_value)
        )

        iv = self.session.generate_random(128)
        crypttext = alice_session.encrypt("HI BOB!", mechanism_param=iv)
        plaintext = bob_session.decrypt(crypttext, mechanism_param=iv)
        self.assertEqual(plaintext, b"HI BOB!")

    @Only.softhsm2
    def test_import_key_params(self):
        # Using explicit curve params is bad practice and many HSMs
        # don't support this usage, so we only test it on SoftHSM
        der = base64.b64decode("""
        MIICXDCCAc8GByqGSM49AgEwggHCAgEBME0GByqGSM49AQECQgH/////////////
        ////////////////////////////////////////////////////////////////
        /////////zCBngRCAf//////////////////////////////////////////////
        ///////////////////////////////////////8BEFRlT65YY4cmh+SmiGgtoVA
        7qLacluZsxXzuLSJkY7xCeFWGTlR7H6TexZSwL07sb8HNXPfiD0sNPHvRR/Ua1A/
        AAMVANCeiAApHLhTlsxnFzkyhKqg2mS6BIGFBADGhY4GtwQE6c2ePstmI5W0Qpxk
        gTkFP7Uh+CivYGtNPbqhS1537+dZKP4dwSei/6jeM0izwYVqQpv5fn4xwuW9ZgEY
        OSlqeJo7wARcil+0LH0b2Zj1RElXm0RoF6+9Fyc+ZiyX7nKZXvQmQMVQuQE/rQdh
        NTxwhqJywkCIvpR2n9FmUAJCAf//////////////////////////////////////
        ////+lGGh4O/L5Zrf8wBSPcJpdA7tcm4iZxHrrtvtx6ROGQJAgEBA4GGAAQBMLgt
        gTFBGr0f7YrWwZsCPpLxaUQvUKvz2C6ghiFmxc2EzBgxDY+ywnmG4T++EVZhJHTP
        eIOnVRcHXXivkRe+YMQBbH/fZyqfCe41vIl39bwhqli839AAj/WoxXZuilpKaXBp
        vGbx2380UIhrec1jFjItOOg/Xp9dOecjQZK7Z0wVq1U=
        """)
        key = self.session.create_object(decode_ec_public_key(der))
        self.assertIsInstance(key, pkcs11.PublicKey)

        # We should get back to identity
        self.assertEqual(encode_ec_public_key(key), der)

    @requires(Mechanism.ECDSA_SHA1)
    def test_import_key_named_curve(self):
        der = base64.b64decode("""
        MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAEa6Q5Hs+j71J1lc+VziafH+uL6603
        R8gTAphQD0iLG9Q9RgAvDQdFFpzkvXI+mEGVNRMmT/BA1OtficHcAXTdXA==
        """)
        key = self.session.create_object(decode_ec_public_key(der))
        self.assertIsInstance(key, pkcs11.PublicKey)

        # Something signed with OpenSSL
        signature = base64.b64decode("""
        MEYCIQD1nDlli+uLuGX3eobKJe7PsRYkYJ4F15bjqbbB+MHewwIhAPGFRwyuFOvH
        zuj+sxXwk1CsDWN7AXbmHufOlOarXpiq
        """)
        signature = decode_ecdsa_signature(signature)

        self.assertTrue(key.verify(b"Data to sign", signature, mechanism=Mechanism.ECDSA_SHA1))

        # We should get back to identity
        self.assertEqual(encode_ec_public_key(key), der)

    @requires(Mechanism.ECDSA)
    def test_import_key_pair(self):
        priv = base64.b64decode("""
        MHcCAQEEIMu1c8rEExH5jAfFy9bIS8RbMoHaKqoyvzrRz5rTUip2oAoGCCqGSM49
        AwEHoUQDQgAEdrKww7nWyfHoT2jqgGK3wFaJGssJJZD0bIY7RsIISqeaT88bU/HK
        44HxKoBkOs/JWHX5m/zrblnz40kjOuPZeA==
        """)
        priv = self.session.create_object(decode_ec_private_key(priv))

        pub = base64.b64decode("""
        MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAEdrKww7nWyfHoT2jqgGK3wFaJGssJ
        JZD0bIY7RsIISqeaT88bU/HK44HxKoBkOs/JWHX5m/zrblnz40kjOuPZeA==
        """)
        pub = self.session.create_object(decode_ec_public_key(pub))

        signature = priv.sign(b"Example", mechanism=Mechanism.ECDSA)
        self.assertTrue(pub.verify(b"Example", signature, mechanism=Mechanism.ECDSA))

    @requires(Mechanism.EC_EDWARDS_KEY_PAIR_GEN, Mechanism.EDDSA)
    def test_sign_ed25519(self):
        parameters = self.session.create_domain_parameters(
            KeyType.EC_EDWARDS,
            {
                Attribute.EC_PARAMS: encode_named_curve_parameters(
                    PrivateKeyAlgorithmId.unmap("ed25519")
                )
            },
            local=True,
        )

        pub, priv = parameters.generate_keypair()

        mechanism = Mechanism.EDDSA
        data = b"HI BOB!"
        eddsa = priv.sign(data, mechanism=mechanism)
        self.assertTrue(pub.verify(data, eddsa, mechanism=mechanism))

    @requires(Mechanism.EC_EDWARDS_KEY_PAIR_GEN, Mechanism.EDDSA)
    def test_sign_ed448(self):
        parameters = self.session.create_domain_parameters(
            KeyType.EC_EDWARDS,
            {
                Attribute.EC_PARAMS: encode_named_curve_parameters(
                    PrivateKeyAlgorithmId.unmap("ed448")
                )
            },
            local=True,
        )

        pub, priv = parameters.generate_keypair()

        mechanism = Mechanism.EDDSA
        data = b"HI BOB!"
        # As per the spec, mechanism parameters are required for Ed448: phFlag is False and
        # the contextData is null for a regular Ed448 signature.
        eddsa = priv.sign(data, mechanism=mechanism, mechanism_param=(False, None))
        self.assertTrue(pub.verify(data, eddsa, mechanism=mechanism, mechanism_param=(False, None)))



from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.ec import ECDSA
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import CBC
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.padding import PKCS7
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    PublicFormat,
    load_der_public_key,
)

from pkcs11 import KDF, Attribute, KeyType, Mechanism, ObjectClass
from pkcs11.util.ec import (
    decode_ec_public_key,
    encode_ec_public_key,
    encode_ecdsa_signature,
    encode_named_curve_parameters,
)
from pkcs11.util.rsa import encode_rsa_public_key

from . import Is, TestCase, requires


class ExternalPublicKeyTests(TestCase):
    @requires(Mechanism.RSA_PKCS)
    def test_rsa(self):
        # A key we generated earlier
        self.session.generate_keypair(KeyType.RSA, 1024)

        pub = self.session.get_key(key_type=KeyType.RSA, object_class=ObjectClass.PUBLIC_KEY)

        pub = encode_rsa_public_key(pub)

        pub = load_der_public_key(pub)
        crypttext = pub.encrypt(b"Data to encrypt", PKCS1v15())

        priv = self.session.get_key(key_type=KeyType.RSA, object_class=ObjectClass.PRIVATE_KEY)

        plaintext = priv.decrypt(crypttext, mechanism=Mechanism.RSA_PKCS)

        self.assertEqual(plaintext, b"Data to encrypt")

    @requires(Mechanism.ECDSA_SHA256)
    def test_ecdsa(self):
        # A key we generated earlier
        self.session.create_domain_parameters(
            KeyType.EC,
            {
                Attribute.EC_PARAMS: encode_named_curve_parameters("secp256r1"),
            },
            local=True,
        ).generate_keypair()

        priv = self.session.get_key(key_type=KeyType.EC, object_class=ObjectClass.PRIVATE_KEY)

        signature = priv.sign(b"Data to sign", mechanism=Mechanism.ECDSA_SHA256)
        signature = encode_ecdsa_signature(signature)

        pub = self.session.get_key(key_type=KeyType.EC, object_class=ObjectClass.PUBLIC_KEY)
        pub = load_der_public_key(encode_ec_public_key(pub))
        pub.verify(signature, b"Data to sign", ECDSA(SHA256()))

    @requires(Mechanism.ECDH1_DERIVE)
    def test_ecdh(self):
        # A key we generated earlier
        self.session.create_domain_parameters(
            KeyType.EC,
            {
                Attribute.EC_PARAMS: encode_named_curve_parameters("secp256r1"),
            },
            local=True,
        ).generate_keypair()

        # Retrieve our keypair, with our public key encoded for interchange
        alice_priv = self.session.get_key(key_type=KeyType.EC, object_class=ObjectClass.PRIVATE_KEY)
        alice_pub = self.session.get_key(key_type=KeyType.EC, object_class=ObjectClass.PUBLIC_KEY)
        alice_pub = encode_ec_public_key(alice_pub)

        # Bob generates a keypair, with their public key encoded for
        # interchange
        bob_priv = ec.generate_private_key(ec.SECP256R1())
        bob_pub = bob_priv.public_key().public_bytes(
            Encoding.DER,
            PublicFormat.SubjectPublicKeyInfo,
        )

        # Bob converts Alice's key to internal format and generates their
        # shared key
        bob_shared_key = bob_priv.exchange(
            ec.ECDH(),
            load_der_public_key(alice_pub),
        )

        key = alice_priv.derive_key(
            KeyType.GENERIC_SECRET,
            256,
            mechanism_param=(
                KDF.NULL,
                None,
                # N.B. it seems like SoftHSMv2 requires an EC_POINT to be
                # DER-encoded, which is not what the spec says
                decode_ec_public_key(bob_pub, encode_ec_point=Is.softhsm2)[Attribute.EC_POINT],
            ),
            template={
                Attribute.SENSITIVE: False,
                Attribute.EXTRACTABLE: True,
            },
        )
        alice_shared_key = key[Attribute.VALUE]

        # We should have the same shared key
        self.assertEqual(bob_shared_key, alice_shared_key)

    @requires(Mechanism.RSA_PKCS)
    def test_terrible_hybrid_file_encryption_app(self):
        # Proof of concept code only!
        import io

        # A key we generated earlier
        self.session.generate_keypair(KeyType.RSA, 1024)

        pub = self.session.get_key(key_type=KeyType.RSA, object_class=ObjectClass.PUBLIC_KEY)
        pub = load_der_public_key(encode_rsa_public_key(pub))

        key = self.session.generate_random(256)
        iv = self.session.generate_random(128)

        source = b"This is my amazing file"

        with io.BytesIO() as dest:
            # Write a 128-byte header containing our key and our IV
            # strictly speaking we don't need to keep the IV secure but
            # we may as well.
            #
            # FIXME: Because this is RSA 1.5, we should fill the rest of the
            # frame with nonsense
            self.assertEqual(dest.write(pub.encrypt(key + iv, PKCS1v15())), 128)

            cipher = Cipher(AES(key), CBC(iv))
            encryptor = cipher.encryptor()
            padder = PKCS7(128).padder()
            padded_data = padder.update(source) + padder.finalize()
            ciphertext = encryptor.update(padded_data) + encryptor.finalize()
            dest.write(ciphertext)

            # Time passes
            dest.seek(0)

            # Look up our private key
            priv = self.session.get_key(key_type=KeyType.RSA, object_class=ObjectClass.PRIVATE_KEY)
            # Read the header
            header = dest.read(priv.key_length // 8)
            header = priv.decrypt(header, mechanism=Mechanism.RSA_PKCS)

            # The first 32 bytes is our key
            key, header = header[:32], header[32:]
            # The next 16 bytes is the IV
            iv = header[:16]
            # We can ignore the rest

            cipher = Cipher(AES(key), CBC(iv))
            decryptor = cipher.decryptor()
            unpadder = PKCS7(128).unpadder()
            padded_plaintext = decryptor.update(dest.read()) + decryptor.finalize()
            plaintext = unpadder.update(padded_plaintext) + unpadder.finalize()

        self.assertEqual(source, plaintext)



"""
PKCS#11 Digests
"""

import hashlib

from pkcs11 import Attribute, KeyType, Mechanism

from . import Not, TestCase, requires


class DigestTests(TestCase):
    @requires(Mechanism.SHA256)
    def test_digest(self):
        data = "THIS IS SOME DATA TO DIGEST"
        digest = self.session.digest(data, mechanism=Mechanism.SHA256)

        self.assertEqual(digest, hashlib.sha256(data.encode("utf-8")).digest())

    @requires(Mechanism.SHA256)
    def test_digest_generator(self):
        data = (
            b"This is ",
            b"some data ",
            b"to digest.",
        )

        digest = self.session.digest(data, mechanism=Mechanism.SHA256)

        m = hashlib.sha256()
        for d in data:
            m.update(d)

        self.assertEqual(digest, m.digest())

    @requires(Mechanism.AES_KEY_GEN, Mechanism.SHA256)
    @Not.nfast  # nFast can't digest keys
    def test_digest_key(self):
        key = self.session.generate_key(
            KeyType.AES,
            128,
            template={
                Attribute.SENSITIVE: False,
                Attribute.EXTRACTABLE: True,
            },
        )

        digest = self.session.digest(key, mechanism=Mechanism.SHA256)

        self.assertEqual(digest, hashlib.sha256(key[Attribute.VALUE]).digest())

    @requires(Mechanism.AES_KEY_GEN, Mechanism.SHA256)
    @Not.nfast  # nFast can't digest keys
    def test_digest_key_data(self):
        key = self.session.generate_key(
            KeyType.AES,
            128,
            template={
                Attribute.SENSITIVE: False,
                Attribute.EXTRACTABLE: True,
            },
        )

        data = (
            b"Some data",
            key,
        )

        digest = self.session.digest(data, mechanism=Mechanism.SHA256)

        m = hashlib.sha256()
        m.update(data[0])
        m.update(data[1][Attribute.VALUE])

        self.assertEqual(digest, m.digest())

    @requires(Mechanism.SHA256)
    def test_digest_stream_interrupt_releases_operation(self):
        data = (
            b"I" * 16,
            b"N" * 16,
            b"P" * 16,
            b"U" * 16,
            b"T" * 10,
        )

        def _data_with_error():
            yield data[0]
            yield data[1]
            yield data[2]
            raise ValueError

        def attempt_digest():
            self.session.digest(_data_with_error(), mechanism=Mechanism.SHA256)

        self.assertRaises(ValueError, attempt_digest)
        # ...try again
        digest = self.session.digest(data, mechanism=Mechanism.SHA256)
        m = hashlib.sha256()
        for d in data:
            m.update(d)
        self.assertEqual(digest, m.digest())



from datetime import datetime
from struct import Struct

from pkcs11.constants import (
    Attribute,
    CertificateType,
    MechanismFlag,
    ObjectClass,
)
from pkcs11.mechanisms import KeyType, Mechanism

# (Pack Function, Unpack Function) functions
handle_bool = (Struct("?").pack, lambda v: False if len(v) == 0 else Struct("?").unpack(v)[0])
handle_ulong = (Struct("L").pack, lambda v: Struct("L").unpack(v)[0])
handle_str = (lambda s: s.encode("utf-8"), lambda b: b.decode("utf-8"))
handle_date = (
    lambda s: s.strftime("%Y%m%d").encode("ascii"),
    lambda s: datetime.strptime(s.decode("ascii"), "%Y%m%d").date(),
)
handle_bytes = (bytes, bytes)
# The PKCS#11 biginteger type is an array of bytes in network byte order.
# If you have an int type, wrap it in biginteger()
handle_biginteger = handle_bytes


def _enum(type_):
    """Factory to pack/unpack ints into IntEnums."""
    pack, unpack = handle_ulong

    return (lambda v: pack(int(v)), lambda v: type_(unpack(v)))


ATTRIBUTE_TYPES = {
    Attribute.ALWAYS_AUTHENTICATE: handle_bool,
    Attribute.ALWAYS_SENSITIVE: handle_bool,
    Attribute.APPLICATION: handle_str,
    Attribute.BASE: handle_biginteger,
    Attribute.CERTIFICATE_TYPE: _enum(CertificateType),
    Attribute.CHECK_VALUE: handle_bytes,
    Attribute.CLASS: _enum(ObjectClass),
    Attribute.COEFFICIENT: handle_biginteger,
    Attribute.DECRYPT: handle_bool,
    Attribute.DERIVE: handle_bool,
    Attribute.EC_PARAMS: handle_bytes,
    Attribute.EC_POINT: handle_bytes,
    Attribute.ENCRYPT: handle_bool,
    Attribute.END_DATE: handle_date,
    Attribute.EXPONENT_1: handle_biginteger,
    Attribute.EXPONENT_2: handle_biginteger,
    Attribute.EXTRACTABLE: handle_bool,
    Attribute.HASH_OF_ISSUER_PUBLIC_KEY: handle_bytes,
    Attribute.HASH_OF_SUBJECT_PUBLIC_KEY: handle_bytes,
    Attribute.ID: handle_bytes,
    Attribute.ISSUER: handle_bytes,
    Attribute.KEY_GEN_MECHANISM: _enum(Mechanism),
    Attribute.KEY_TYPE: _enum(KeyType),
    Attribute.LABEL: handle_str,
    Attribute.LOCAL: handle_bool,
    Attribute.MODIFIABLE: handle_bool,
    Attribute.COPYABLE: handle_bool,
    Attribute.MODULUS: handle_biginteger,
    Attribute.MODULUS_BITS: handle_ulong,
    Attribute.NEVER_EXTRACTABLE: handle_bool,
    Attribute.OBJECT_ID: handle_bytes,
    Attribute.PRIME: handle_biginteger,
    Attribute.PRIME_BITS: handle_ulong,
    Attribute.PRIME_1: handle_biginteger,
    Attribute.PRIME_2: handle_biginteger,
    Attribute.PRIVATE: handle_bool,
    Attribute.PRIVATE_EXPONENT: handle_biginteger,
    Attribute.PUBLIC_EXPONENT: handle_biginteger,
    Attribute.SENSITIVE: handle_bool,
    Attribute.SERIAL_NUMBER: handle_bytes,
    Attribute.SIGN: handle_bool,
    Attribute.SIGN_RECOVER: handle_bool,
    Attribute.START_DATE: handle_date,
    Attribute.SUBJECT: handle_bytes,
    Attribute.SUBPRIME: handle_biginteger,
    Attribute.SUBPRIME_BITS: handle_ulong,
    Attribute.TOKEN: handle_bool,
    Attribute.TRUSTED: handle_bool,
    Attribute.UNIQUE_ID: handle_str,
    Attribute.UNWRAP: handle_bool,
    Attribute.URL: handle_str,
    Attribute.VALUE: handle_biginteger,
    Attribute.VALUE_BITS: handle_ulong,
    Attribute.VALUE_LEN: handle_ulong,
    Attribute.VERIFY: handle_bool,
    Attribute.VERIFY_RECOVER: handle_bool,
    Attribute.WRAP: handle_bool,
    Attribute.WRAP_WITH_TRUSTED: handle_bool,
    Attribute.GOSTR3410_PARAMS: handle_bytes,
    Attribute.GOSTR3411_PARAMS: handle_bytes,
}
"""
Map of attributes to (serialize, deserialize) functions.
"""

ALL_CAPABILITIES = (
    Attribute.ENCRYPT,
    Attribute.DECRYPT,
    Attribute.WRAP,
    Attribute.UNWRAP,
    Attribute.SIGN,
    Attribute.VERIFY,
    Attribute.DERIVE,
)


def _apply_common(template, id_, label, store):
    if id_:
        template[Attribute.ID] = id_
    if label:
        template[Attribute.LABEL] = label
    template[Attribute.TOKEN] = bool(store)


def _apply_capabilities(template, possible_capas, capabilities):
    for attr in possible_capas:
        template[attr] = _capa_attr_to_mechanism_flag[attr] & capabilities


_capa_attr_to_mechanism_flag = {
    Attribute.ENCRYPT: MechanismFlag.ENCRYPT,
    Attribute.DECRYPT: MechanismFlag.DECRYPT,
    Attribute.WRAP: MechanismFlag.WRAP,
    Attribute.UNWRAP: MechanismFlag.UNWRAP,
    Attribute.SIGN: MechanismFlag.SIGN,
    Attribute.VERIFY: MechanismFlag.VERIFY,
    Attribute.DERIVE: MechanismFlag.DERIVE,
}


class AttributeMapper:
    """
    Class mapping PKCS#11 attributes to and from Python values.
    """

    def __init__(self):
        self.attribute_types = dict(ATTRIBUTE_TYPES)
        self.default_secret_key_template = {
            Attribute.CLASS: ObjectClass.SECRET_KEY,
            Attribute.ID: b"",
            Attribute.LABEL: "",
            Attribute.PRIVATE: True,
            Attribute.SENSITIVE: True,
        }
        self.default_public_key_template = {
            Attribute.CLASS: ObjectClass.PUBLIC_KEY,
            Attribute.ID: b"",
            Attribute.LABEL: "",
        }
        self.default_private_key_template = {
            Attribute.CLASS: ObjectClass.PRIVATE_KEY,
            Attribute.ID: b"",
            Attribute.LABEL: "",
            Attribute.PRIVATE: True,
            Attribute.SENSITIVE: True,
        }

    def register_handler(self, key, pack, unpack):
        self.attribute_types[key] = (pack, unpack)

    def _handler(self, key):
        try:
            return self.attribute_types[key]
        except KeyError as e:
            raise NotImplementedError(f"Can't handle attribute type {hex(key)}.") from e

    def pack_attribute(self, key, value):
        """Pack a Attribute value into a bytes array."""
        pack, _ = self._handler(key)
        return pack(value)

    def unpack_attributes(self, key, value):
        """Unpack a Attribute bytes array into a Python value."""
        _, unpack = self._handler(key)
        return unpack(value)

    def public_key_template(
        self,
        *,
        capabilities,
        id_,
        label,
        store,
    ):
        template = self.default_public_key_template
        _apply_capabilities(
            template, (Attribute.ENCRYPT, Attribute.WRAP, Attribute.VERIFY), capabilities
        )
        _apply_common(template, id_, label, store)
        return template

    def private_key_template(
        self,
        *,
        capabilities,
        id_,
        label,
        store,
    ):
        template = self.default_private_key_template
        _apply_capabilities(
            template,
            (Attribute.DECRYPT, Attribute.UNWRAP, Attribute.SIGN, Attribute.DERIVE),
            capabilities,
        )
        _apply_common(template, id_, label, store)
        return template

    def secret_key_template(
        self,
        *,
        capabilities,
        id_,
        label,
        store,
    ):
        return self.generic_key_template(
            self.default_secret_key_template,
            capabilities=capabilities,
            id_=id_,
            label=label,
            store=store,
        )

    def generic_key_template(
        self,
        base_template,
        *,
        capabilities,
        id_,
        label,
        store,
    ):
        template = dict(base_template)
        _apply_capabilities(template, ALL_CAPABILITIES, capabilities)
        _apply_common(template, id_, label, store)
        return template



"""
X.509 Certificate Tests
"""

import base64
import datetime

from asn1crypto import pem
from asn1crypto.csr import CertificationRequest, CertificationRequestInfo
from asn1crypto.keys import RSAPublicKey
from asn1crypto.x509 import Certificate, Name, TbsCertificate, Time
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.serialization import load_der_public_key

import pkcs11
from pkcs11 import (
    Attribute,
    KeyType,
    Mechanism,
)
from pkcs11.util.dsa import decode_dsa_signature
from pkcs11.util.ec import decode_ecdsa_signature
from pkcs11.util.rsa import encode_rsa_public_key
from pkcs11.util.x509 import decode_x509_certificate, decode_x509_public_key

from . import Not, TestCase, requires

# X.509 self-signed certificate (generated with OpenSSL)
# openssl req -x509 \
#   -newkey rsa:512 \
#   -keyout key.pem \
#   -out cert.pem \
#   -days 365 \
#   -nodes
_, _, CERT = pem.unarmor(b"""
-----BEGIN CERTIFICATE-----
MIICKzCCAdWgAwIBAgIJAK3BO9rnLZd9MA0GCSqGSIb3DQEBBQUAMEUxCzAJBgNV
BAYTAkFVMRMwEQYDVQQIEwpTb21lLVN0YXRlMSEwHwYDVQQKExhJbnRlcm5ldCBX
aWRnaXRzIFB0eSBMdGQwHhcNMTcwNjAyMDI0ODMyWhcNMTgwNjAyMDI0ODMyWjBF
MQswCQYDVQQGEwJBVTETMBEGA1UECBMKU29tZS1TdGF0ZTEhMB8GA1UEChMYSW50
ZXJuZXQgV2lkZ2l0cyBQdHkgTHRkMFwwDQYJKoZIhvcNAQEBBQADSwAwSAJBAK5z
DJiUDIutdWY8sT2O2ABKh5nmWjc4uEjNj/i5ZLQ4YlRmDL4e2vWs/GOFLVtTJKj6
rh4fj65Xo6X/5R/y+U8CAwEAAaOBpzCBpDAdBgNVHQ4EFgQU+cG240Pzz0y6igtm
hnk1+1KFv6gwdQYDVR0jBG4wbIAU+cG240Pzz0y6igtmhnk1+1KFv6ihSaRHMEUx
CzAJBgNVBAYTAkFVMRMwEQYDVQQIEwpTb21lLVN0YXRlMSEwHwYDVQQKExhJbnRl
cm5ldCBXaWRnaXRzIFB0eSBMdGSCCQCtwTva5y2XfTAMBgNVHRMEBTADAQH/MA0G
CSqGSIb3DQEBBQUAA0EAOdvMKLrIFOYF3aVLGharY196heO0fndm39sZAXJ4PItx
n28DytHEdAoltksfJ2Ds3XAjQqcpI5eBbhIoN9Ckxg==
-----END CERTIFICATE-----
""")


class X509Tests(TestCase):
    def test_import_ca_certificate_easy(self):
        cert = self.session.create_object(decode_x509_certificate(CERT))
        self.assertIsInstance(cert, pkcs11.Certificate)

    @Not.nfast
    @Not.opencryptoki
    def test_import_ca_certificate(self):
        cert = self.session.create_object(decode_x509_certificate(CERT, extended_set=True))
        self.assertIsInstance(cert, pkcs11.Certificate)

        self.assertEqual(
            cert[Attribute.HASH_OF_ISSUER_PUBLIC_KEY],
            b"\xf9\xc1\xb6\xe3\x43\xf3\xcf\x4c\xba\x8a\x0b\x66\x86\x79\x35\xfb\x52\x85\xbf\xa8",
        )
        # Cert is self signed
        self.assertEqual(
            cert[Attribute.HASH_OF_SUBJECT_PUBLIC_KEY],
            b"\xf9\xc1\xb6\xe3\x43\xf3\xcf\x4c\xba\x8a\x0b\x66\x86\x79\x35\xfb\x52\x85\xbf\xa8",
        )

    @requires(Mechanism.SHA1_RSA_PKCS)
    def test_verify_certificate_rsa(self):
        # Warning: proof of concept code only!
        x509 = Certificate.load(CERT)
        key = self.session.create_object(decode_x509_public_key(CERT))
        self.assertIsInstance(key, pkcs11.PublicKey)

        value = x509["tbs_certificate"].dump()
        signature = x509.signature

        assert x509.signature_algo == "rsassa_pkcs1v15"
        assert x509.hash_algo == "sha1"

        self.assertTrue(key.verify(value, signature, mechanism=Mechanism.SHA1_RSA_PKCS))

    @requires(Mechanism.DSA_SHA1)
    def test_verify_certificate_dsa(self):
        # Warning: proof of concept code only!
        CERT = base64.b64decode("""
        MIIDbjCCAy6gAwIBAgIJAKPBInGiPjXNMAkGByqGSM44BAMwRTELMAkGA1UEBhMC
        QVUxEzARBgNVBAgTClNvbWUtU3RhdGUxITAfBgNVBAoTGEludGVybmV0IFdpZGdp
        dHMgUHR5IEx0ZDAeFw0xNzA3MDMxMjI1MTBaFw0xOTA3MDMxMjI1MTBaMEUxCzAJ
        BgNVBAYTAkFVMRMwEQYDVQQIEwpTb21lLVN0YXRlMSEwHwYDVQQKExhJbnRlcm5l
        dCBXaWRnaXRzIFB0eSBMdGQwggG3MIIBLAYHKoZIzjgEATCCAR8CgYEA7U0AshA/
        4MXQ3MHykoeotEoPc+OXFMJ2PHzKfbFD80UC5bloxC9kp908GG3emdqbJuCTfVUD
        sex1vEgMj1sEwilBow954zMqncu5lLBIGZKjT6tloW8sFt50sE0l+YnBvAiw9uoL
        9lBOZLKh87zWPZUuORm8lWhZEwjUnZ+3S5ECFQCNJGd68RpctgkA1kDp33NhQhev
        lQKBgQCQ6uYkvNpHMtXwyGII4JyOyStbteHjHdKfJfLNRyIEEq/E4e3Do6NGIr26
        Z7u9iBsA5/aU6gKSBrYprxY1hdR4gTRBNzSUDEzf7IX3bfRIbBhjlNBSBba5Fs0z
        /kszZbZ8XYGVxs92aWFk/1JIZ0wnToC794+juq72/TvrtvxdowOBhAACgYAjoknQ
        kRD0+x3GkbngQCU+VNspZuXboB22CU3bDGVAVhmI5N02M8NmeuN7SqqYZAlw01Ju
        rzBF7i9VW4qxBaWszMCwyozerSVjZ2JA/Qubb57v/p7F3FDHq7E33FZzgyhOimds
        rzXpVErCGJJ1oBGz5H5fvoKnQmfh0X8N/VHkZqOBpzCBpDAdBgNVHQ4EFgQUQayv
        usUnpvRgc9OtXGddqMiwm5cwdQYDVR0jBG4wbIAUQayvusUnpvRgc9OtXGddqMiw
        m5ehSaRHMEUxCzAJBgNVBAYTAkFVMRMwEQYDVQQIEwpTb21lLVN0YXRlMSEwHwYD
        VQQKExhJbnRlcm5ldCBXaWRnaXRzIFB0eSBMdGSCCQCjwSJxoj41zTAMBgNVHRME
        BTADAQH/MAkGByqGSM44BAMDLwAwLAIUNE+zTuFe01v0BRTLarPtGK8ZHHcCFB9Y
        YAwtpblAgUEdGuoAtnoEQ2tc
        """)

        x509 = Certificate.load(CERT)
        key = self.session.create_object(decode_x509_public_key(CERT))
        self.assertIsInstance(key, pkcs11.PublicKey)

        value = x509["tbs_certificate"].dump()

        assert x509.signature_algo == "dsa"
        assert x509.hash_algo == "sha1"

        signature = decode_dsa_signature(x509.signature)

        self.assertTrue(key.verify(value, signature, mechanism=Mechanism.DSA_SHA1))

    @requires(Mechanism.ECDSA_SHA256)
    def test_verify_certificate_ecdsa(self):
        # Warning: proof of concept code only!
        CERT = base64.b64decode("""
        MIICgzCCAgmgAwIBAgICEAAwCgYIKoZIzj0EAwIwUTELMAkGA1UEBhMCQkUxFDAS
        BgNVBAoMC0V4YW1wbGUgSW5jMRowGAYDVQQLDBFUZXN0aW5nIEF1dGhvcml0eTEQ
        MA4GA1UEAwwHUm9vdCBDQTAgFw0wMDAxMDEwMDAwMDBaGA8yNTAwMDEwMTAwMDAw
        MFowUTELMAkGA1UEBhMCQkUxFDASBgNVBAoMC0V4YW1wbGUgSW5jMRowGAYDVQQL
        DBFUZXN0aW5nIEF1dGhvcml0eTEQMA4GA1UEAwwHUm9vdCBDQTB2MBAGByqGSM49
        AgEGBSuBBAAiA2IABIxRV+HCT5hbggdOa0CxbOyLRgCRQIFHnsjwk7UZCBeb+SHb
        r4zHM447nASOEwJKvc37UttkdC4lpdOjw9OkwltCSMCS2s22v18//MqjRoQ8wAiX
        hk1mR499ltu1jKicDKOBsTCBrjAdBgNVHQ4EFgQUJNkIpnJ27yAJidmTShDvCLfz
        PJYwHwYDVR0jBBgwFoAUJNkIpnJ27yAJidmTShDvCLfzPJYwDwYDVR0TAQH/BAUw
        AwEB/zAOBgNVHQ8BAf8EBAMCAYYwSwYDVR0fBEQwQjBAoD6gPIY6aHR0cDovL3B5
        aGFua28udGVzdHMvdGVzdGluZy1jYS1lY2RzYS9jcmxzL3Jvb3QvbGF0ZXN0LmNy
        bDAKBggqhkjOPQQDAgNoADBlAjApktbaE81Qil3bbI5UFWqpH4JsW1pgucZTlQN+
        VmXMRT/0SVHTMM64IK1B8CzVhI8CMQCFbdX+K7KZYNDYuA7gTQHdp7l12PXMoBGE
        dcda0K/1qwvA2w6mNU1qi/b0Is7oA0I=
        """)

        x509 = Certificate.load(CERT)
        key = self.session.create_object(decode_x509_public_key(CERT))
        self.assertIsInstance(key, pkcs11.PublicKey)

        value = x509["tbs_certificate"].dump()

        assert x509.signature_algo == "ecdsa"
        assert x509.hash_algo == "sha256"

        signature = decode_ecdsa_signature(x509.signature)

        self.assertTrue(key.verify(value, signature, mechanism=Mechanism.ECDSA_SHA256))

    @requires(Mechanism.RSA_PKCS_KEY_PAIR_GEN, Mechanism.SHA256_RSA_PKCS)
    def test_self_sign_certificate(self):
        # Warning: proof of concept code only!
        pub, priv = self.session.generate_keypair(KeyType.RSA, 1024)
        pub_asn1 = RSAPublicKey.load(encode_rsa_public_key(pub))

        tbs = TbsCertificate(
            {
                "version": "v1",
                "serial_number": 1,
                "issuer": Name.build(
                    {
                        "common_name": "Test Certificate",
                    }
                ),
                "subject": Name.build(
                    {
                        "common_name": "Test Certificate",
                    }
                ),
                "signature": {
                    "algorithm": "sha256_rsa",
                    "parameters": None,
                },
                "validity": {
                    "not_before": Time(
                        {
                            "utc_time": datetime.datetime(
                                2017, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
                            ),
                        }
                    ),
                    "not_after": Time(
                        {
                            "utc_time": datetime.datetime(
                                2038, 12, 31, 23, 59, tzinfo=datetime.timezone.utc
                            ),
                        }
                    ),
                },
                "subject_public_key_info": {
                    "algorithm": {
                        "algorithm": "rsa",
                        "parameters": None,
                    },
                    "public_key": pub_asn1,
                },
            }
        )

        # Sign the TBS Certificate
        value = priv.sign(tbs.dump(), mechanism=Mechanism.SHA256_RSA_PKCS)

        cert = Certificate(
            {
                "tbs_certificate": tbs,
                "signature_algorithm": {
                    "algorithm": "sha256_rsa",
                    "parameters": None,
                },
                "signature_value": value,
            }
        )

        der_cert = cert.dump()

        # read back the data and validate it
        pub_key_handle = load_der_public_key(pub_asn1.dump())
        cert_loaded = Certificate.load(der_cert)
        tbs_bytes = cert_loaded["tbs_certificate"].dump()
        signature_bytes = cert_loaded["signature_value"].native
        pub_key_handle.verify(signature_bytes, tbs_bytes, PKCS1v15(), SHA256())

    @requires(Mechanism.RSA_PKCS_KEY_PAIR_GEN, Mechanism.SHA256_RSA_PKCS)
    def test_sign_csr(self):
        # Warning: proof of concept code only!
        pub, priv = self.session.generate_keypair(KeyType.RSA, 1024)

        pub_asn1 = RSAPublicKey.load(encode_rsa_public_key(pub))

        info = CertificationRequestInfo(
            {
                "version": 0,
                "subject": Name.build(
                    {
                        "common_name": "Test Certificate",
                    }
                ),
                "subject_pk_info": {
                    "algorithm": {
                        "algorithm": "rsa",
                        "parameters": None,
                    },
                    "public_key": pub_asn1,
                },
            }
        )

        # Sign the CSR Info
        value = priv.sign(info.dump(), mechanism=Mechanism.SHA256_RSA_PKCS)

        csr_data = CertificationRequest(
            {
                "certification_request_info": info,
                "signature_algorithm": {
                    "algorithm": "sha256_rsa",
                    "parameters": None,
                },
                "signature": value,
            }
        ).dump()

        # read back the data and validate it
        pub_key_handle = load_der_public_key(pub_asn1.dump())
        csr_loaded = CertificationRequest.load(csr_data)
        info_bytes = csr_loaded["certification_request_info"].dump()
        signature_bytes = csr_loaded["signature"].native
        pub_key_handle.verify(signature_bytes, info_bytes, PKCS1v15(), SHA256())



"""
Iterator tests
"""

import pkcs11

from . import TestCase, requires


class IteratorTests(TestCase):
    @requires(pkcs11.Mechanism.AES_KEY_GEN, pkcs11.Mechanism.AES_CBC_PAD)
    def test_partial_decrypt(self):
        self.session.generate_key(pkcs11.KeyType.AES, 128, label="LOOK ME UP")

        key = self.session.get_key(label="LOOK ME UP")
        data = (
            b"1234",
            b"1234",
        )

        iv = self.session.generate_random(128)
        encrypted_data = list(key.encrypt(data, mechanism_param=iv))

        iter1 = key.decrypt(encrypted_data, mechanism_param=iv)
        next(iter1)

        with self.assertRaises(pkcs11.OperationActive):
            iter2 = key.decrypt(encrypted_data, mechanism_param=iv)
            next(iter2)

    @requires(pkcs11.Mechanism.AES_KEY_GEN, pkcs11.Mechanism.AES_CBC_PAD)
    def test_close_iterators(self):
        self.session.generate_key(pkcs11.KeyType.AES, 128, label="LOOK ME UP")

        key = self.session.get_key(label="LOOK ME UP")
        data = (
            b"1234",
            b"1234",
        )

        iv = self.session.generate_random(128)
        encrypted_data = list(key.encrypt(data, mechanism_param=iv))

        iter1 = key.decrypt(encrypted_data, mechanism_param=iv)
        next(iter1)
        del iter1

        iter2 = key.decrypt(encrypted_data, mechanism_param=iv)
        next(iter2)
